import os
import sys
import tempfile
import unittest

import osc.core as osc_core
from osc.store import Store


class TestStore(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='osc_test')
        self.store = Store(self.tmpdir, check=False)
        self.store.write_string("_osclib_version", Store.STORE_VERSION)
        self.store.apiurl = "http://localhost"
        self.store.is_package = True
        self.store.project = "project name"
        self.store.package = "package name"

    def tearDown(self):
        try:
            shutil.rmtree(self.tmpdir)
        except:
            pass

    def fileEquals(self, fn, expected_value):
        path = os.path.join(self.tmpdir, ".osc", fn)
        with open(path) as f:
            actual_value = f.read()
            self.assertEqual(actual_value, expected_value, f"File: {fn}")

    def test_read_write_file(self):
        self.store.write_file("_file", "\n\nline1\nline2")
        self.fileEquals("_file", "\n\nline1\nline2")
        self.assertEqual(self.store.read_file("_file"), "\n\nline1\nline2")

        # writing None removes the file
        self.store.write_file("_file", None)
        self.assertFalse(self.store.exists("_file"))

        self.assertRaises(TypeError, self.store.write_string, "_file", 123)
        self.assertRaises(TypeError, self.store.write_string, "_file", ["123"])

    def test_read_write_int(self):
        self.store.write_int("_int", 123)
        self.fileEquals("_int", "123\n")
        self.assertEqual(self.store.read_int("_int"), 123)

        # writing None removes the file
        self.store.write_int("_int", None)
        self.assertFalse(self.store.exists("_int"))

        self.assertRaises(TypeError, self.store.write_int, "_int", "123")
        self.assertRaises(TypeError, self.store.write_int, "_int", b"123")
        self.assertRaises(TypeError, self.store.write_int, "_int", ["123"])

    def test_read_write_list(self):
        self.store.write_list("_list", ["one", "two", "three"])
        self.fileEquals("_list", "one\ntwo\nthree\n")
        self.assertEqual(self.store.read_list("_list"), ["one", "two", "three"])

        # writing None removes the file
        self.store.write_list("_list", None)
        self.assertFalse(self.store.exists("_list"))

        self.assertRaises(TypeError, self.store.write_list, "_list", "123")
        self.assertRaises(TypeError, self.store.write_list, "_list", b"123")
        self.assertRaises(TypeError, self.store.write_list, "_list", 123)

    def test_read_write_string(self):
        self.store.write_string("_string", "string")
        self.fileEquals("_string", "string\n")
        self.assertEqual(self.store.read_string("_string"), "string")

        self.store.write_string("_bytes", b"bytes")
        self.fileEquals("_bytes", "bytes\n")
        self.assertEqual(self.store.read_string("_bytes"), "bytes")

        # writing None removes the file
        self.store.write_string("_string", None)
        self.assertFalse(self.store.exists("_string"))

        self.assertRaises(TypeError, self.store.write_string, "_string", 123)
        self.assertRaises(TypeError, self.store.write_string, "_string", ["123"])

    def test_contains(self):
        self.assertTrue("_project" in self.store)
        self.assertTrue("_package" in self.store)
        self.assertFalse("_foo" in self.store)

    def test_iter(self):
        self.assertEqual(len(list(self.store)), 4)
        for fn in self.store:
            self.assertIn(fn, ["_osclib_version", "_apiurl", "_project", "_package"])

    def test_apiurl(self):
        self.store.apiurl = "https://example.com"
        self.fileEquals("_apiurl", "https://example.com\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.apiurl, "https://example.com")

    def test_apiurl_no_trailing_slash(self):
        self.store.apiurl = "https://example.com/"
        self.fileEquals("_apiurl", "https://example.com\n")

        self.store.write_string("_apiurl", "https://example.com/")
        self.fileEquals("_apiurl", "https://example.com/\n")
        self.assertEqual(self.store.apiurl, "https://example.com")

    def test_package(self):
        self.fileEquals("_package", "package name\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.package, "package name")

    def test_project(self):
        self.fileEquals("_project", "project name\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.project, "project name")

    def test_scmurl(self):
        self.store.scmurl = "https://example.com/project.git"
        self.fileEquals("_scm", "https://example.com/project.git\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.scmurl, "https://example.com/project.git")

    def test_size_limit(self):
        self.store.size_limit = 123
        self.fileEquals("_size_limit", "123\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.size_limit, 123)

    def test_to_be_added(self):
        self.store.to_be_added = ["foo", "bar", "baz"]
        self.fileEquals("_to_be_added", "foo\nbar\nbaz\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.to_be_added, ["foo", "bar", "baz"])

    def test_to_be_deleted(self):
        self.store.to_be_deleted = ["foo", "bar", "baz"]
        self.fileEquals("_to_be_deleted", "foo\nbar\nbaz\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.to_be_deleted, ["foo", "bar", "baz"])

    def test_in_conflict(self):
        self.store.in_conflict = ["foo", "bar", "baz"]
        self.fileEquals("_in_conflict", "foo\nbar\nbaz\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.in_conflict, ["foo", "bar", "baz"])

    def test_osclib_version(self):
        # no setter, users are not supposed to set the version
        self.assertRaises(AttributeError, setattr, self.store, "osclib_version", "123")
        self.store.write_string("_osclib_version", "123")
        self.fileEquals("_osclib_version", "123\n")

        store2 = Store(self.tmpdir, check=False)
        self.assertEqual(store2.osclib_version, "123")

    def test_files(self):
        files = [
            osc_core.File(name="foo", md5="aabbcc", size=1, mtime=2),
            osc_core.File(name="bar", md5="ddeeff", size=3, mtime=4, skipped=True),
        ]

        self.store.files = files

        expected = """
<directory>
  <entry name="bar" md5="ddeeff" size="3" mtime="4" skipped="true" />
  <entry name="foo" md5="aabbcc" size="1" mtime="2" />
</directory>""".lstrip()

        if sys.version_info[:2] <= (3, 7):
            # ElementTree doesn't preserve attribute order on py <= 3.7; https://bugs.python.org/issue34160
            expected = """
<directory>
  <entry md5="ddeeff" mtime="4" name="bar" size="3" skipped="true" />
  <entry md5="aabbcc" mtime="2" name="foo" size="1" />
</directory>""".lstrip()

        self.fileEquals("_files", expected)

        store2 = Store(self.tmpdir)
        files2 = store2.files
        # files got ordered
        self.assertTrue(files2[0] == files[1])
        self.assertTrue(files2[1] == files[0])

    def test_last_buildroot(self):
        self.assertEqual(self.store.last_buildroot, None)

        self.store.last_buildroot = "repo", "arch", "vm_type"
        self.fileEquals("_last_buildroot", "repo\narch\nvm_type\n")

        self.assertRaises(ValueError, setattr, self.store, "last_buildroot", ["one"])
        self.assertRaises(ValueError, setattr, self.store, "last_buildroot", ["one", "two"])
        self.assertRaises(ValueError, setattr, self.store, "last_buildroot", ["one", "two", "three", "four"])

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.last_buildroot, ["repo", "arch", "vm_type"])

        self.store.last_buildroot = "repo", "arch", None
        self.fileEquals("_last_buildroot", "repo\narch\n\n")

        store2 = Store(self.tmpdir)
        self.assertEqual(store2.last_buildroot, ["repo", "arch", None])

    def test_meta_node(self):
        self.store.write_string(
            "_meta",
            """<package name="test-pkgA" project="projectA">
  <title>title</title>
  <description>desc</description>
  <releasename>name</releasename>
  <build>
    <enable repository="repo1"/>
    <enable repository="repo2"/>
  </build>
</package>""",
        )
        node = self.store._meta_node
        self.assertNotEqual(node, None)

        # try to read the _meta via a package class
        from osc._private import LocalPackage

        self.store.files = []
        pkg = LocalPackage(self.tmpdir)
        self.assertEqual(pkg.get_meta_value("releasename"), "name")

    def test_migrate_10_20_sources_file(self):
        self.store = Store(self.tmpdir, check=False)
        self.store.write_string("_osclib_version", "1.0")
        self.store.apiurl = "http://localhost"
        self.store.is_package = True
        self.store.project = "project name"
        self.store.package = "package name"
        self.store.write_string("sources", "")
        self.store.files = [
            osc_core.File(name="sources", md5="aabbcc", size=0, mtime=0),
        ]

        Store(self.tmpdir, check=True)
        self.assertTrue(os.path.exists(os.path.join(self.tmpdir, ".osc", "sources", "sources")))


if __name__ == "__main__":
    unittest.main()



import os
import re
import struct
import sys
from . import packagequery

from .helper import decode_it


def cmp(a, b):
    return (a > b) - (a < b)


class RpmError(packagequery.PackageError):
    pass


class RpmHeaderError(RpmError):
    pass


class RpmHeader:
    """corresponds more or less to the indexEntry_s struct"""

    def __init__(self, offset, length):
        self.offset = offset
        # length of the data section (without length of indexEntries)
        self.length = length
        self.entries = []

    def append(self, entry):
        self.entries.append(entry)

    def gettag(self, tag):
        for i in self.entries:
            if i.tag == tag:
                return i
        return None

    def __iter__(self):
        yield from self.entries

    def __len__(self):
        return len(self.entries)


class RpmHeaderEntry:
    """corresponds to the entryInfo_s struct (except the data attribute)"""

    # each element represents an int
    ENTRY_SIZE = 16

    def __init__(self, tag, type, offset, count):
        self.tag = tag
        self.type = type
        self.offset = offset
        self.count = count
        self.data = None


class RpmQuery(packagequery.PackageQuery, packagequery.PackageQueryResult):
    LEAD_SIZE = 96
    LEAD_MAGIC = 0xedabeedb
    HEADER_MAGIC = 0x8eade801
    HEADERSIG_TYPE = 5

    LESS = 1 << 1
    GREATER = 1 << 2
    EQUAL = 1 << 3

    SENSE_STRONG = 1 << 27

    default_tags = (
        1000, 1001, 1002, 1003, 1004, 1022, 1005, 1020,
        1047, 1112, 1113,  # provides
        1049, 1048, 1050,  # requires
        1054, 1053, 1055,  # conflicts
        1090, 1114, 1115,  # obsoletes
        1156, 1158, 1157,  # oldsuggests
        5046, 5047, 5048,  # recommends
        5049, 5051, 5050,  # suggests
        5052, 5053, 5054,  # supplements
        5055, 5056, 5057  # enhances
    )

    def __init__(self, fh):
        self.__file = fh
        self.__path = os.path.abspath(fh.name)
        self.filename_suffix = 'rpm'
        self.header = None

    def read(self, all_tags=False, self_provides=True, *extra_tags, **extra_kw):
        # self_provides is unused because a rpm always has a self provides
        self.__read_lead()
        data = self.__file.read(RpmHeaderEntry.ENTRY_SIZE)
        hdrmgc, reserved, il, dl = struct.unpack('!I3i', data)
        if self.HEADER_MAGIC != hdrmgc:
            raise RpmHeaderError(self.__path, 'invalid headermagic \'%s\'' % hdrmgc)
        # skip signature header for now
        size = il * RpmHeaderEntry.ENTRY_SIZE + dl
        # data is 8 byte aligned
        pad = (size + 7) & ~7
        querysig = extra_kw.get('querysig')
        if not querysig:
            self.__file.read(pad)
            data = self.__file.read(RpmHeaderEntry.ENTRY_SIZE)
        hdrmgc, reserved, il, dl = struct.unpack('!I3i', data)
        self.header = RpmHeader(pad, dl)
        if self.HEADER_MAGIC != hdrmgc:
            raise RpmHeaderError(self.__path, 'invalid headermagic \'%s\'' % hdrmgc)
        data = self.__file.read(il * RpmHeaderEntry.ENTRY_SIZE)
        while len(data) > 0:
            ei = struct.unpack('!4i', data[:RpmHeaderEntry.ENTRY_SIZE])
            self.header.append(RpmHeaderEntry(*ei))
            data = data[RpmHeaderEntry.ENTRY_SIZE:]
        data = self.__file.read(self.header.length)
        for i in self.header:
            if i.tag in self.default_tags + extra_tags or all_tags:
                try:  # this may fail for -debug* packages
                    self.__read_data(i, data)
                except:
                    pass
        return self

    def __read_lead(self):
        data = self.__file.read(self.LEAD_SIZE)
        leadmgc, = struct.unpack('!I', data[:4])
        if leadmgc != self.LEAD_MAGIC:
            raise RpmError(self.__path, 'not a rpm (invalid lead magic \'%s\')' % leadmgc)
        sigtype, = struct.unpack('!h', data[78:80])
        if sigtype != self.HEADERSIG_TYPE:
            raise RpmError(self.__path, 'invalid header signature \'%s\'' % sigtype)

    def __read_data(self, entry, data):
        off = entry.offset
        if entry.type == 2:
            entry.data = struct.unpack('!%dc' % entry.count, data[off:off + 1 * entry.count])
        if entry.type == 3:
            entry.data = struct.unpack('!%dh' % entry.count, data[off:off + 2 * entry.count])
        elif entry.type == 4:
            entry.data = struct.unpack('!%di' % entry.count, data[off:off + 4 * entry.count])
        elif entry.type == 6:
            entry.data = unpack_string(data[off:])
        elif entry.type == 7:
            entry.data = data[off:off + entry.count]
        elif entry.type == 8 or entry.type == 9:
            cnt = entry.count
            entry.data = []
            while cnt > 0:
                cnt -= 1
                s = unpack_string(data[off:])
                # also skip '\0'
                off += len(s) + 1
                entry.data.append(s)
            if entry.type == 8:
                return
            lang = os.getenv('LANGUAGE') or os.getenv('LC_ALL') \
                or os.getenv('LC_MESSAGES') or os.getenv('LANG')
            if lang is None:
                entry.data = entry.data[0]
                return
            # get private i18n table
            table = self.header.gettag(100)
            # just care about the country code
            lang = lang.split('_', 1)[0]
            cnt = 0
            for i in table.data:
                if cnt > len(entry.data) - 1:
                    break
                if i == lang:
                    entry.data = entry.data[cnt]
                    return
                cnt += 1
            entry.data = entry.data[0]
        else:
            raise RpmHeaderError(self.__path, 'unsupported tag type \'%d\' (tag: \'%s\'' % (entry.type, entry.tag))

    def __reqprov(self, tag, flags, version, strong=None):
        pnames = self.header.gettag(tag)
        if not pnames:
            return []
        pnames = pnames.data
        pflags = self.header.gettag(flags).data
        pvers = self.header.gettag(version).data
        if not (pnames and pflags and pvers):
            raise RpmError(self.__path, 'cannot get provides/requires, tags are missing')
        res = []
        for name, flags, ver in zip(pnames, pflags, pvers):
            if strong is not None:
                # compat code for the obsolete RPMTAG_OLDSUGGESTSNAME tag
                # strong == 1 => return only "recommends"
                # strong == 0 => return only "suggests"
                if strong == 1:
                    strong = self.SENSE_STRONG
                if (flags & self.SENSE_STRONG) != strong:
                    continue
            # RPMSENSE_SENSEMASK = 15 (see rpmlib.h) but ignore RPMSENSE_SERIAL (= 1 << 0) therefore use 14
            if flags & 14:
                name += b' '
                if flags & self.GREATER:
                    name += b'>'
                elif flags & self.LESS:
                    name += b'<'
                if flags & self.EQUAL:
                    name += b'='
                name += b' %s' % ver
            res.append(name)
        return res

    def vercmp(self, rpmq):
        res = RpmQuery.rpmvercmp(str(self.epoch()), str(rpmq.epoch()))
        if res != 0:
            return res
        res = RpmQuery.rpmvercmp(self.version(), rpmq.version())
        if res != 0:
            return res
        res = RpmQuery.rpmvercmp(self.release(), rpmq.release())
        return res

    # XXX: create dict for the tag => number mapping?!
    def name(self):
        return self.header.gettag(1000).data

    def version(self):
        return self.header.gettag(1001).data

    def release(self):
        return self.header.gettag(1002).data

    def epoch(self):
        epoch = self.header.gettag(1003)
        if epoch is None:
            return 0
        return epoch.data[0]

    def arch(self):
        return self.header.gettag(1022).data

    def summary(self):
        return self.header.gettag(1004).data

    def description(self):
        return self.header.gettag(1005).data

    def url(self):
        entry = self.header.gettag(1020)
        if entry is None:
            return None
        return entry.data

    def path(self):
        return self.__path

    def provides(self):
        return self.__reqprov(1047, 1112, 1113)

    def requires(self):
        return self.__reqprov(1049, 1048, 1050)

    def conflicts(self):
        return self.__reqprov(1054, 1053, 1055)

    def obsoletes(self):
        return self.__reqprov(1090, 1114, 1115)

    def recommends(self):
        recommends = self.__reqprov(5046, 5048, 5047)
        if not recommends:
            recommends = self.__reqprov(1156, 1158, 1157, 1)
        return recommends

    def suggests(self):
        suggests = self.__reqprov(5049, 5051, 5050)
        if not suggests:
            suggests = self.__reqprov(1156, 1158, 1157, 0)
        return suggests

    def supplements(self):
        return self.__reqprov(5052, 5054, 5053)

    def enhances(self):
        return self.__reqprov(5055, 5057, 5506)

    def is_src(self):
        # SOURCERPM = 1044
        return self.gettag(1044) is None

    def is_nosrc(self):
        # NOSOURCE = 1051, NOPATCH = 1052
        return self.is_src() and \
            (self.gettag(1051) is not None or self.gettag(1052) is not None)

    def gettag(self, num):
        return self.header.gettag(num)

    def canonname(self):
        if self.is_nosrc():
            arch = b'nosrc'
        elif self.is_src():
            arch = b'src'
        else:
            arch = self.arch()
        return RpmQuery.filename(self.name(), None, self.version(), self.release(), arch)

    @staticmethod
    def query(filename):
        f = open(filename, 'rb')
        rpmq = RpmQuery(f)
        rpmq.read()
        f.close()
        return rpmq

    @staticmethod
    def queryhdrmd5(filename):
        f = open(filename, 'rb')
        rpmq = RpmQuery(f)
        rpmq.read(1004, querysig=True)
        f.close()
        entry = rpmq.gettag(1004)
        if entry is None:
            return None
        return ''.join(["%02x" % x for x in struct.unpack('16B', entry.data)])

    @staticmethod
    def rpmvercmp(ver1, ver2):
        """
        implementation of RPM's version comparison algorithm
        (as described in lib/rpmvercmp.c)
        """
        if ver1 == ver2:
            return 0
        res = 0
        ver1 = decode_it(ver1)
        ver2 = decode_it(ver2)
        while res == 0:
            # remove all leading non alphanumeric or tilde chars
            ver1 = re.sub('^[^a-zA-Z0-9~]*', '', ver1)
            ver2 = re.sub('^[^a-zA-Z0-9~]*', '', ver2)
            if ver1.startswith('~') or ver2.startswith('~'):
                if not ver1.startswith('~'):
                    return 1
                elif not ver2.startswith('~'):
                    return -1
                ver1 = ver1[1:]
                ver2 = ver2[1:]
                continue

            if not (len(ver1) and len(ver2)):
                break

            # check if we have a digits segment
            mo1 = re.match(r'(\d+)', ver1)
            mo2 = re.match(r'(\d+)', ver2)
            numeric = True
            if mo1 is None:
                mo1 = re.match('([a-zA-Z]+)', ver1)
                mo2 = re.match('([a-zA-Z]+)', ver2)
                numeric = False
            # check for different types: alpha and numeric
            if mo2 is None:
                if numeric:
                    return 1
                return -1
            seg1 = mo1.group(0)
            ver1 = ver1[mo1.end(0):]
            seg2 = mo2.group(1)
            ver2 = ver2[mo2.end(1):]
            if numeric:
                # remove leading zeros
                seg1 = re.sub('^0+', '', seg1)
                seg2 = re.sub('^0+', '', seg2)
                # longer digit segment wins - if both have the same length
                # a simple ascii compare decides
                res = len(seg1) - len(seg2) or cmp(seg1, seg2)
            else:
                res = cmp(seg1, seg2)
        if res > 0:
            return 1
        elif res < 0:
            return -1
        return cmp(ver1, ver2)

    @staticmethod
    def filename(name, epoch, version, release, arch):
        return b'%s-%s-%s.%s.rpm' % (name, version, release, arch)


def unpack_string(data, encoding=None):
    """unpack a '\\0' terminated string from data"""
    idx = data.find(b'\0')
    if idx == -1:
        raise ValueError('illegal string: not \\0 terminated')
    data = data[:idx]
    if encoding is not None:
        data = data.decode(encoding)
    return data


if __name__ == '__main__':
    try:
        rpmq = RpmQuery.query(sys.argv[1])
    except RpmError as e:
        print(e.msg)
        sys.exit(2)
    print(rpmq.name(), rpmq.version(), rpmq.release(), rpmq.arch(), rpmq.url())
    print(rpmq.summary())
    print(rpmq.description())
    print('##########')
    print('\n'.join(rpmq.provides()))
    print('##########')
    print('\n'.join(rpmq.requires()))
    print('##########')
    print(RpmQuery.queryhdrmd5(sys.argv[1]))



import argparse
import os
import shutil
import tempfile
import unittest

from osc.commandline import Command
from osc.commandline import MainCommand
from osc.commandline import OscMainCommand
from osc.commandline import pop_project_package_from_args
from osc.commandline import pop_project_package_repository_arch_from_args
from osc.commandline import pop_project_package_targetproject_targetpackage_from_args
from osc.commandline import pop_repository_arch_from_args
from osc.oscerr import NoWorkingCopy, OscValueError
from osc.store import Store


class TestMainCommand(MainCommand):
    name = "osc-test"

    def init_arguments(self, command=None):
        self.add_argument(
            "-A",
            "--apiurl",
        )


class TestCommand(Command):
    name = "test-cmd"


OSCRC_LOCALHOST = """
[general]
apiurl = https://localhost

[https://localhost]
user=Admin
pass=opensuse
""".lstrip()


class TestCommandClasses(unittest.TestCase):
    def setUp(self):
        os.environ.pop("OSC_CONFIG", None)
        self.tmpdir = tempfile.mkdtemp(prefix="osc_test")
        os.chdir(self.tmpdir)
        self.oscrc = None

    def tearDown(self):
        os.environ.pop("OSC_CONFIG", None)
        try:
            shutil.rmtree(self.tmpdir)
        except OSError:
            pass

    def write_oscrc_localhost(self):
        self.oscrc = os.path.join(self.tmpdir, "oscrc")
        with open(self.oscrc, "w") as f:
            f.write(OSCRC_LOCALHOST)

    def test_load_commands(self):
        main = TestMainCommand()
        main.load_commands()

    def test_load_command(self):
        main = TestMainCommand()
        cmd = main.load_command(TestCommand, "test.osc.commands")
        self.assertTrue(str(cmd).startswith("<osc plugin test.osc.commands.TestCommand"))

    def test_parent(self):
        class Parent(TestCommand):
            name = "parent"

        class Child(TestCommand):
            name = "child"
            parent = "Parent"

        main = TestMainCommand()
        main.load_command(Parent, "test.osc.commands")
        main.load_command(Child, "test.osc.commands")

        main.parse_args(["parent", "child"])

    def test_invalid_parent(self):
        class Parent(TestCommand):
            name = "parent"

        class Child(TestCommand):
            name = "child"
            parent = "DoesNotExist"

        main = TestMainCommand()
        main.load_command(Parent, "test.osc.commands")
        main.load_command(Child, "test.osc.commands")

    def test_load_twice(self):
        class AnotherCommand(TestCommand):
            name = "another-command"
            aliases = ["test-cmd"]

        main = TestMainCommand()
        main.load_command(TestCommand, "test.osc.commands")

        # conflict between names
        self.assertRaises(argparse.ArgumentError, main.load_command, TestCommand, "test.osc.commands")

        # conflict between a name and an alias
        self.assertRaises(argparse.ArgumentError, main.load_command, AnotherCommand, "test.osc.commands")

    def test_intermixing(self):
        main = TestMainCommand()
        main.load_command(TestCommand, "test.osc.commands")

        args = main.parse_args(["test-cmd", "--apiurl", "https://example.com"])
        self.assertEqual(args.apiurl, "https://example.com")

        args = main.parse_args(["--apiurl", "https://example.com", "test-cmd"])
        self.assertEqual(args.apiurl, "https://example.com")

    def test_unknown_options(self):
        main = TestMainCommand()
        main.load_command(TestCommand, "test.osc.commands")

        args = main.parse_args(["test-cmd", "unknown-arg"])
        self.assertEqual(args.positional_args, ["unknown-arg"])

        self.assertRaises(SystemExit, main.parse_args, ["test-cmd", "--unknown-option"])

    def test_default_apiurl(self):
        class TestMainCommand(OscMainCommand):
            name = "osc-test"

        main = TestMainCommand()
        main.load_command(TestCommand, "test.osc.commands")

        self.write_oscrc_localhost()
        os.environ["OSC_CONFIG"] = self.oscrc
        args = main.parse_args(["test-cmd"])
        main.post_parse_args(args)
        self.assertEqual(args.apiurl, "https://localhost")


class TestPopProjectPackageFromArgs(unittest.TestCase):
    def _write_store(self, project=None, package=None):
        store = Store(self.tmpdir, check=False)
        store.write_string("_osclib_version", Store.STORE_VERSION)
        store.apiurl = "http://localhost"
        if project:
            store.project = project
            store.is_project = True
        if package:
            store.package = package
            store.is_project = False
            store.is_package = True

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix="osc_test")
        os.chdir(self.tmpdir)

    def tearDown(self):
        try:
            shutil.rmtree(self.tmpdir)
        except OSError:
            pass

    def test_individual_args(self):
        args = ["project", "package", "another-arg"]
        project, package = pop_project_package_from_args(args)
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(args, ["another-arg"])

    def test_slash_separator(self):
        args = ["project/package", "another-arg"]
        project, package = pop_project_package_from_args(args)
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(args, ["another-arg"])

        args = ["project/", "another-arg"]
        project, package = pop_project_package_from_args(args)
        self.assertEqual(project, "project")
        self.assertEqual(package, "")
        self.assertEqual(args, ["another-arg"])

        args = ["/", "another-arg"]
        project, package = pop_project_package_from_args(args)
        self.assertEqual(project, "")
        self.assertEqual(package, "")
        self.assertEqual(args, ["another-arg"])

    def test_missing_project(self):
        args = []
        self.assertRaises(
            OscValueError,
            pop_project_package_from_args,
            args,
        )

    def test_optional_project(self):
        args = []
        project, package = pop_project_package_from_args(
            args, project_is_optional=True,
        )
        self.assertEqual(project, None)
        self.assertEqual(package, None)
        self.assertEqual(args, [])

    def test_default_project(self):
        args = []
        project, package = pop_project_package_from_args(
            args, default_project="default-project", package_is_optional=True
        )
        self.assertEqual(project, "default-project")
        self.assertEqual(package, None)
        self.assertEqual(args, [])

    def test_missing_package(self):
        args = ["project"]
        self.assertRaises(
            OscValueError,
            pop_project_package_from_args,
            args,
        )

    def test_optional_package(self):
        args = ["project"]
        project, package = pop_project_package_from_args(
            args, package_is_optional=True,
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, None)
        self.assertEqual(args, [])

    def test_default_package(self):
        args = ["project"]
        project, package = pop_project_package_from_args(
            args, package_is_optional=True, default_package="default-package",
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, None)
        self.assertEqual(args, [])

    def test_default_project_package(self):
        args = []
        project, package = pop_project_package_from_args(
            args,
            default_project="default-project",
            default_package="default-package",
            package_is_optional=True,
        )
        self.assertEqual(project, "default-project")
        self.assertEqual(package, "default-package")
        self.assertEqual(args, [])

    def test_no_working_copy(self):
        args = [".", "."]
        self.assertRaises(NoWorkingCopy, pop_project_package_from_args, args)

        args = [".", "package"]
        self.assertRaises(NoWorkingCopy, pop_project_package_from_args, args)

        args = ["project", "."]
        self.assertRaises(NoWorkingCopy, pop_project_package_from_args, args)

    def test_project_and_package_from_project_working_copy(self):
        self._write_store("store_project")

        args = [".", ".", "another-arg"]
        self.assertRaises(NoWorkingCopy, pop_project_package_from_args, args)

        args = ["."]
        project, package = pop_project_package_from_args(args, package_is_optional=True)
        self.assertEqual(project, "store_project")
        self.assertEqual(package, None)
        self.assertEqual(args, [])

        args = []
        self.assertRaises(
            NoWorkingCopy,
            pop_project_package_from_args,
            args,
            default_project=".",
            default_package=".",
        )

        args = []
        project, package = pop_project_package_from_args(
            args, default_project=".", default_package=".", package_is_optional=True
        )
        self.assertEqual(project, "store_project")
        self.assertEqual(package, None)
        self.assertEqual(args, [])

    def test_project_and_package_from_package_working_copy(self):
        self._write_store("store_project", "store_package")

        args = [".", ".", "another-arg"]
        project, package = pop_project_package_from_args(args)
        self.assertEqual(project, "store_project")
        self.assertEqual(package, "store_package")
        self.assertEqual(args, ["another-arg"])

        args = ["."]
        project, package = pop_project_package_from_args(args, package_is_optional=True)
        self.assertEqual(project, "store_project")
        self.assertEqual(package, None)
        self.assertEqual(args, [])

        args = []
        project, package = pop_project_package_from_args(
            args, default_project=".", default_package="."
        )
        self.assertEqual(project, "store_project")
        self.assertEqual(package, "store_package")
        self.assertEqual(args, [])

        args = []
        project, package = pop_project_package_from_args(
            args, default_project=".", default_package=".", package_is_optional=True
        )
        self.assertEqual(project, "store_project")
        self.assertEqual(package, "store_package")
        self.assertEqual(args, [])

    def test_project_requires_to_specify_package(self):
        args = ["project"]
        self.assertRaises(
            OscValueError,
            pop_project_package_from_args,
            args,
            default_project=".",
            default_package=".",
        )

        # The project from store is ignored because we've specified one.
        # Specifying a package is expected.
        self._write_store("store_project")
        args = ["project"]
        self.assertRaises(
            OscValueError,
            pop_project_package_from_args,
            args,
            default_project=".",
            default_package=".",
        )


class TestPopRepositoryArchFromArgs(unittest.TestCase):
    def test_individial_args(self):
        args = ["repo", "arch", "another-arg"]
        repo, arch = pop_repository_arch_from_args(args)
        self.assertEqual(repo, "repo")
        self.assertEqual(arch, "arch")
        self.assertEqual(args, ["another-arg"])

    def test_slash_separator(self):
        args = ["repo/arch", "another-arg"]
        repo, arch = pop_repository_arch_from_args(args)
        self.assertEqual(repo, "repo")
        self.assertEqual(arch, "arch")
        self.assertEqual(args, ["another-arg"])

        args = ["repo/", "another-arg"]
        repo, arch = pop_repository_arch_from_args(args)
        self.assertEqual(repo, "repo")
        self.assertEqual(arch, "")
        self.assertEqual(args, ["another-arg"])

        args = ["/", "another-arg"]
        repo, arch = pop_repository_arch_from_args(args)
        self.assertEqual(repo, "")
        self.assertEqual(arch, "")
        self.assertEqual(args, ["another-arg"])

    def test_missing_repository(self):
        args = []
        self.assertRaises(OscValueError, pop_repository_arch_from_args, args)

    def test_optional_repository(self):
        args = []
        repo, arch = pop_repository_arch_from_args(args, repository_is_optional=True)
        self.assertEqual(repo, None)
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_default_repository(self):
        args = []
        repo, arch = pop_repository_arch_from_args(args, default_repository="repo", arch_is_optional=True)
        self.assertEqual(repo, "repo")
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_missing_arch(self):
        args = ["repo"]
        self.assertRaises(OscValueError, pop_repository_arch_from_args, args)

    def test_optional_arch(self):
        args = ["repo"]
        repo, arch = pop_repository_arch_from_args(args, arch_is_optional=True)
        self.assertEqual(repo, "repo")
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_default_arch(self):
        args = ["repo"]
        self.assertRaises(OscValueError, pop_repository_arch_from_args, args, default_arch="arch")


class TestPopProjectPackageRepositoryArchFromArgs(unittest.TestCase):
    def _write_store(self, project=None, package=None):
        store = Store(self.tmpdir, check=False)
        store.write_string("_osclib_version", Store.STORE_VERSION)
        store.apiurl = "http://localhost"
        if project:
            store.project = project
            store.is_project = True
        if package:
            store.package = package
            store.is_project = False
            store.is_package = True

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix="osc_test")
        os.chdir(self.tmpdir)

    def tearDown(self):
        try:
            shutil.rmtree(self.tmpdir)
        except OSError:
            pass

    def test_individual_args(self):
        args = ["project", "package", "repo", "arch", "another-arg"]
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(repo, "repo")
        self.assertEqual(arch, "arch")
        self.assertEqual(args, ["another-arg"])

    def test_slash_separator(self):
        args = ["project/package", "repo/arch", "another-arg"]
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(repo, "repo")
        self.assertEqual(arch, "arch")
        self.assertEqual(args, ["another-arg"])

    def test_missing_project(self):
        args = []
        self.assertRaises(
            OscValueError, pop_project_package_repository_arch_from_args, args
        )

    def test_optional_project(self):
        args = []
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args, project_is_optional=True
        )
        self.assertEqual(project, None)
        self.assertEqual(package, None)
        self.assertEqual(repo, None)
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_default_project(self):
        args = []
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args, default_project="project", package_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, None)
        self.assertEqual(repo, None)
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_missing_package(self):
        args = ["project"]
        self.assertRaises(
            OscValueError, pop_project_package_repository_arch_from_args, args
        )

    def test_optional_package(self):
        args = ["project"]
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args, package_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, None)
        self.assertEqual(repo, None)
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_default_package(self):
        args = ["project"]
        self.assertRaises(
            OscValueError,
            pop_project_package_repository_arch_from_args,
            args, default_package="package",
            repository_is_optional=True,
        )

    def test_missing_repository(self):
        args = ["project", "package"]
        self.assertRaises(
            OscValueError, pop_project_package_repository_arch_from_args, args
        )

    def test_optional_repository(self):
        args = ["project", "package"]
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args, repository_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(repo, None)
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_default_repository(self):
        args = ["project", "package"]
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args, default_repository="repository", arch_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(repo, "repository")
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_missing_arch(self):
        args = ["project", "package", "repo"]
        self.assertRaises(
            OscValueError, pop_project_package_repository_arch_from_args, args
        )

    def test_optional_arch(self):
        args = ["project", "package", "repository"]
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args, arch_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(repo, "repository")
        self.assertEqual(arch, None)
        self.assertEqual(args, [])

    def test_default_arch(self):
        args = ["project", "package", "repository"]
        self.assertRaises(
            OscValueError,
            pop_project_package_repository_arch_from_args,
            args,
            default_arch="arch",
        )

    def test_no_working_copy(self):
        args = ["repo", "arch"]
        self.assertRaises(
            NoWorkingCopy,
            pop_project_package_repository_arch_from_args,
            args,
            default_project=".",
            default_package=".",
        )

    def test_working_copy(self):
        self._write_store("store_project", "store_package")
        args = ["repo", "arch"]
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args,
            default_project=".",
            default_package=".",
        )
        self.assertEqual(project, "store_project")
        self.assertEqual(package, "store_package")
        self.assertEqual(repo, "repo")
        self.assertEqual(arch, "arch")

    def test_working_copy_extra_arg(self):
        self._write_store("store_project", "store_package")
        args = ["repo", "arch", "another-arg"]
        # example of invalid usage, working copy is not used when there's 3+ args; [project, package, ...] are expected
        self.assertRaises(
            OscValueError,
            pop_project_package_repository_arch_from_args,
            args,
            default_project=".",
            default_package=".",
        )

    def test_working_copy_optional_arch(self):
        self._write_store("store_project", "store_package")
        args = ["repository"]
        project, package, repo, arch = pop_project_package_repository_arch_from_args(
            args,
            default_project=".",
            default_package=".",
            arch_is_optional=True,
        )
        self.assertEqual(project, "store_project")
        self.assertEqual(package, "store_package")
        self.assertEqual(repo, "repository")
        self.assertEqual(arch, None)


class TestPopProjectPackageTargetProjectTargetPackageFromArgs(unittest.TestCase):
    def _write_store(self, project=None, package=None):
        store = Store(self.tmpdir, check=False)
        store.write_string("_osclib_version", Store.STORE_VERSION)
        store.apiurl = "http://localhost"
        if project:
            store.project = project
            store.is_project = True
        if package:
            store.package = package
            store.is_project = False
            store.is_package = True

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix="osc_test")
        os.chdir(self.tmpdir)

    def tearDown(self):
        try:
            shutil.rmtree(self.tmpdir)
        except OSError:
            pass

    def test_individual_args(self):
        args = ["project", "package", "target-project", "target-package", "another-arg"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(target_project, "target-project")
        self.assertEqual(target_package, "target-package")
        self.assertEqual(args, ["another-arg"])

    def test_slash_separator(self):
        args = ["project/package", "target-project/target-package", "another-arg"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(target_project, "target-project")
        self.assertEqual(target_package, "target-package")
        self.assertEqual(args, ["another-arg"])

    def test_missing_project(self):
        args = []
        self.assertRaises(
            OscValueError, pop_project_package_targetproject_targetpackage_from_args, args
        )

    def test_optional_project(self):
        args = []
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args, project_is_optional=True
        )
        self.assertEqual(project, None)
        self.assertEqual(package, None)
        self.assertEqual(target_project, None)
        self.assertEqual(target_package, None)
        self.assertEqual(args, [])

    def test_default_project(self):
        args = []
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args, default_project="project", package_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, None)
        self.assertEqual(target_project, None)
        self.assertEqual(target_package, None)
        self.assertEqual(args, [])

    def test_missing_package(self):
        args = ["project"]
        self.assertRaises(
            OscValueError, pop_project_package_targetproject_targetpackage_from_args, args
        )

    def test_optional_package(self):
        args = ["project"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args, package_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, None)
        self.assertEqual(target_project, None)
        self.assertEqual(target_package, None)
        self.assertEqual(args, [])

    def test_default_package(self):
        args = ["project"]
        self.assertRaises(
            OscValueError,
            pop_project_package_targetproject_targetpackage_from_args,
            args, default_package="package",
            target_project_is_optional=True,
        )

    def test_missing_target_project(self):
        args = ["project", "package"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args, target_project_is_optional=True, target_package_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(target_project, None)
        self.assertEqual(target_package, None)
        self.assertEqual(args, [])

    def test_optional_target_project(self):
        args = ["project", "package"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args, target_project_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(target_project, None)
        self.assertEqual(target_package, None)
        self.assertEqual(args, [])

    def test_default_target_project(self):
        args = ["project", "package"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args, default_target_project="package", target_package_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(target_project, "package")
        self.assertEqual(target_package, None)
        self.assertEqual(args, [])

    def test_missing_target_package(self):
        args = ["project", "package", "target-project"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args, target_package_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(target_project, "target-project")
        self.assertEqual(target_package, None)
        self.assertEqual(args, [])

    def test_optional_target_package(self):
        args = ["project", "package", "target-project"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args, target_package_is_optional=True
        )
        self.assertEqual(project, "project")
        self.assertEqual(package, "package")
        self.assertEqual(target_project, "target-project")
        self.assertEqual(target_package, None)
        self.assertEqual(args, [])

    def test_default_target_package(self):
        args = ["project", "package", "target-project"]

        self.assertRaises(
            OscValueError,
            pop_project_package_targetproject_targetpackage_from_args,
            args,
            default_target_package="target-package",
        )

    def test_no_working_copy(self):
        args = ["target-project", "target-package"]
        self.assertRaises(
            NoWorkingCopy,
            pop_project_package_targetproject_targetpackage_from_args,
            args,
            default_project=".",
            default_package=".",
        )

    def test_working_copy(self):
        self._write_store("store_project", "store_package")
        args = ["target-project", "target-package"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args,
            default_project=".",
            default_package=".",
        )
        self.assertEqual(project, "store_project")
        self.assertEqual(package, "store_package")
        self.assertEqual(target_project, "target-project")
        self.assertEqual(target_package, "target-package")

    def test_working_copy_missing_target_package(self):
        self._write_store("store_project", "store_package")
        args = ["target-project"]
        project, package, target_project, target_package = pop_project_package_targetproject_targetpackage_from_args(
            args,
            default_project=".",
            default_package=".",
            target_package_is_optional=True,
        )
        self.assertEqual(project, "store_project")
        self.assertEqual(package, "store_package")
        self.assertEqual(target_project, "target-project")
        self.assertEqual(target_package, None)

    def test_working_copy_extra_arg(self):
        self._write_store("store_project", "store_package")
        args = ["target-project", "target-package", "another-arg"]
        # example of invalid usage, working copy is not used when there's 3+ args; [project, package, ...] are expected
        self.assertRaises(
            OscValueError,
            pop_project_package_targetproject_targetpackage_from_args,
            args,
        )


if __name__ == "__main__":
    unittest.main()



import osc.commandline


class RepoListCommand(osc.commandline.OscCommand):
    """
    List repositories in project meta
    """

    name = "list"
    aliases = ["ls"]
    parent = "RepoCommand"

    def init_arguments(self):
        self.add_argument(
            "project",
            help="Name of the project",
        )

    def run(self, args):
        from .. import obs_api
        from ..output import KeyValueTable

        project_obj = obs_api.Project.from_api(args.apiurl, args.project)
        repo_flags = project_obj.resolve_repository_flags()

        flag_map = {}
        for (repo_name, arch), data in repo_flags.items():
            for flag_name, flag_value in data.items():
                if flag_value is None:
                    continue
                action = "enable" if flag_value else "disable"
                flag_map.setdefault(repo_name, {}).setdefault(flag_name, {}).setdefault(action, []).append(arch)

        table = KeyValueTable()
        for repo in project_obj.repository_list or []:
            table.add("Repository", repo.name, color="bold")
            table.add("Architectures", ", ".join(repo.arch_list))
            if repo.path_list:
                paths = [f"{path.project}/{path.repository}" for path in repo.path_list]
                table.add("Paths", paths)

            if repo.name in flag_map:
                table.add("Flags", None)
                for flag_name in flag_map[repo.name]:
                    lines = []
                    for action, archs in flag_map[repo.name][flag_name].items():
                        lines.append(f"{action + ':':<8s} {', '.join(archs)}")
                    lines.sort()
                    table.add(flag_name, lines, indent=4)

            table.newline()
        print(str(table))



"""Module for reading repodata directory (created with createrepo) for package
information instead of scanning individual rpms."""


import gzip
import os
from xml.etree import ElementTree as ET

from . import rpmquery
from . import packagequery


def namespace(name):
    return "{http://linux.duke.edu/metadata/%s}" % name


OPERATOR_BY_FLAGS = {
    "EQ": "=",
    "LE": "<=",
    "GE": ">=",
    "LT": "<",
    "GT": ">"
}


def primaryPath(directory):
    """Returns path to the primary repository data file.

    :param directory: repository directory that contains the repodata subdirectory
    :return:  path to primary repository data file
    :rtype: str
    :raise IOError: if repomd.xml contains no primary location
    """
    from .xml import xml_parse

    metaDataPath = os.path.join(directory, "repodata", "repomd.xml")
    elementTree = xml_parse(metaDataPath)
    root = elementTree.getroot()

    for dataElement in root:
        if dataElement.get("type") == "primary":
            locationElement = dataElement.find(namespace("repo") + "location")
            # even though the repomd.xml file is under repodata, the location a
            # attribute is relative to parent directory (directory).
            primaryPath = os.path.join(directory, locationElement.get("href"))
            break
    else:
        raise OSError("'%s' contains no primary location" % metaDataPath)

    return primaryPath


def queries(directory):
    """Returns a list of RepoDataQueries constructed from the repodata under
    the directory.

    :param directory: path to a repository directory (parent directory of repodata directory)
    :return: list of RepoDataQueryResult instances
    :raise IOError: if repomd.xml contains no primary location
    """
    from .xml import xml_parse

    path = primaryPath(directory)

    gunzippedPrimary = gzip.GzipFile(path)
    elementTree = xml_parse(gunzippedPrimary)
    root = elementTree.getroot()

    packageQueries = []
    for packageElement in root:
        packageQuery = RepoDataQueryResult(directory, packageElement)
        packageQueries.append(packageQuery)

    return packageQueries


def _to_bytes_or_None(method):
    def _method(self, *args, **kwargs):
        res = method(self, *args, **kwargs)
        if res is None:
            return None
        return res.encode()

    return _method


def _to_bytes_list(method):
    def _method(self, *args, **kwargs):
        res = method(self, *args, **kwargs)
        return [data.encode() for data in res]

    return _method


class RepoDataQueryResult(packagequery.PackageQueryResult):
    """PackageQueryResult that reads in data from the repodata directory files."""

    def __init__(self, directory, element):
        """Creates a RepoDataQueryResult from the a package Element under a metadata
        Element in a primary.xml file.

        :param directory: repository directory path. Used to convert relative paths to full paths.
        :param element: package Element
        """
        self.__directory = os.path.abspath(directory)
        self.__element = element

    def __formatElement(self):
        return self.__element.find(namespace("common") + "format")

    def __parseEntry(self, element):
        entry = element.get("name")
        flags = element.get("flags")

        if flags is not None:
            version = element.get("ver")
            operator = OPERATOR_BY_FLAGS[flags]
            entry += " %s %s" % (operator, version)

            release = element.get("rel")
            if release is not None:
                entry += "-%s" % release

        return entry

    def __parseEntryCollection(self, collection):
        formatElement = self.__formatElement()
        collectionElement = formatElement.find(namespace("rpm") + collection)

        entries = []
        if collectionElement is not None:
            for entryElement in collectionElement.findall(namespace("rpm") + "entry"):
                entry = self.__parseEntry(entryElement)
                entries.append(entry)

        return entries

    def __versionElement(self):
        return self.__element.find(namespace("common") + "version")

    @_to_bytes_or_None
    def arch(self):
        return self.__element.findtext(namespace("common") + "arch")

    @_to_bytes_or_None
    def description(self):
        return self.__element.findtext(namespace("common") + "description")

    def distribution(self):
        return None

    @_to_bytes_or_None
    def epoch(self):
        return self.__versionElement().get("epoch")

    @_to_bytes_or_None
    def name(self):
        return self.__element.findtext(namespace("common") + "name")

    def path(self):
        locationElement = self.__element.find(namespace("common") + "location")
        relativePath = locationElement.get("href")
        absolutePath = os.path.join(self.__directory, relativePath)

        return absolutePath

    @_to_bytes_list
    def provides(self):
        return self.__parseEntryCollection("provides")

    @_to_bytes_or_None
    def release(self):
        return self.__versionElement().get("rel")

    @_to_bytes_list
    def requires(self):
        return self.__parseEntryCollection("requires")

    @_to_bytes_list
    def conflicts(self):
        return self.__parseEntryCollection('conflicts')

    @_to_bytes_list
    def obsoletes(self):
        return self.__parseEntryCollection('obsoletes')

    @_to_bytes_list
    def recommends(self):
        return self.__parseEntryCollection('recommends')

    @_to_bytes_list
    def suggests(self):
        return self.__parseEntryCollection('suggests')

    @_to_bytes_list
    def supplements(self):
        return self.__parseEntryCollection('supplements')

    @_to_bytes_list
    def enhances(self):
        return self.__parseEntryCollection('enhances')

    def canonname(self):
        if self.release() is None:
            release = None
        else:
            release = self.release()
        return rpmquery.RpmQuery.filename(self.name(), None, self.version(), release, self.arch())

    def gettag(self, tag):
        # implement me, if needed
        return None

    def vercmp(self, other):
        # if either self.epoch() or other.epoch() is None, the vercmp will do
        # the correct thing because one is transformed into b'None' and the
        # other one into b"b'<epoch>'" (and 'b' is greater than 'N')
        res = rpmquery.RpmQuery.rpmvercmp(str(self.epoch()).encode(), str(other.epoch()).encode())
        if res != 0:
            return res
        res = rpmquery.RpmQuery.rpmvercmp(self.version(), other.version())
        if res != 0:
            return res
        res = rpmquery.RpmQuery.rpmvercmp(self.release(), other.release())
        return res

    @_to_bytes_or_None
    def version(self):
        return self.__versionElement().get("ver")



import os
import re
import subprocess
import sys

from . import packagequery


class ArchError(packagequery.PackageError):
    pass


class ArchQuery(packagequery.PackageQuery, packagequery.PackageQueryResult):
    def __init__(self, fh):
        self.__file = fh
        self.__path = os.path.abspath(fh.name)
        self.fields = {}
        # self.magic = None
        # self.pkgsuffix = 'pkg.tar.gz'
        self.pkgsuffix = b'arch'

    def read(self, all_tags=True, self_provides=True, *extra_tags):
        # all_tags and *extra_tags are currently ignored
        f = open(self.__path, 'rb')
        # self.magic = f.read(5)
        # if self.magic == '\375\067zXZ':
        #    self.pkgsuffix = 'pkg.tar.xz'
        fn = open('/dev/null', 'wb')
        pipe = subprocess.Popen(['tar', '-O', '-xf', self.__path, '.PKGINFO'], stdout=subprocess.PIPE, stderr=fn).stdout
        for line in pipe.readlines():
            line = line.rstrip().split(b' = ', 2)
            if len(line) == 2:
                field, value = line[0].decode('ascii'), line[1]
                self.fields.setdefault(field, []).append(value)
        if self_provides:
            prv = b'%s = %s' % (self.name(), self.fields['pkgver'][0])
            self.fields.setdefault('provides', []).append(prv)
        return self

    def vercmp(self, archq):
        res = packagequery.cmp(int(self.epoch()), int(archq.epoch()))
        if res != 0:
            return res
        res = ArchQuery.rpmvercmp(self.version(), archq.version())
        if res != 0:
            return res
        res = ArchQuery.rpmvercmp(self.release(), archq.release())
        return res

    def name(self):
        return self.fields['pkgname'][0] if 'pkgname' in self.fields else None

    def version(self):
        pkgver = self.fields['pkgver'][0] if 'pkgver' in self.fields else None
        if pkgver is not None:
            pkgver = re.sub(br'[0-9]+:', b'', pkgver, 1)
            pkgver = re.sub(br'-[^-]*$', b'', pkgver)
        return pkgver

    def release(self):
        pkgver = self.fields['pkgver'][0] if 'pkgver' in self.fields else None
        if pkgver is not None:
            m = re.search(br'-([^-])*$', pkgver)
            if m:
                return m.group(1)
        return None

    def _epoch(self):
        pkgver = self.fields.get('pkgver', [b''])[0]
        if pkgver:
            m = re.match(br'([0-9])+:', pkgver)
            if m:
                return m.group(1)
        return b''

    def epoch(self):
        epoch = self._epoch()
        if epoch:
            return epoch
        return b'0'

    def arch(self):
        return self.fields['arch'][0] if 'arch' in self.fields else None

    def description(self):
        return self.fields['pkgdesc'][0] if 'pkgdesc' in self.fields else None

    def path(self):
        return self.__path

    def provides(self):
        return self.fields['provides'] if 'provides' in self.fields else []

    def requires(self):
        return self.fields['depend'] if 'depend' in self.fields else []

    def conflicts(self):
        return self.fields['conflict'] if 'conflict' in self.fields else []

    def obsoletes(self):
        return self.fields['replaces'] if 'replaces' in self.fields else []

    def recommends(self):
        # a .PKGINFO has no notion of "recommends"
        return []

    def suggests(self):
        # libsolv treats an optdepend as a "suggests", hence we do the same
        if 'optdepend' not in self.fields:
            return []
        return [re.sub(b':.*', b'', entry) for entry in self.fields['optdepend']]

    def supplements(self):
        # a .PKGINFO has no notion of "recommends"
        return []

    def enhances(self):
        # a .PKGINFO has no notion of "enhances"
        return []

    def canonname(self):
        name = self.name()
        if name is None:
            raise ArchError(self.path(), 'package has no name')
        version = self.version()
        if version is None:
            raise ArchError(self.path(), 'package has no version')
        arch = self.arch()
        if arch is None:
            raise ArchError(self.path(), 'package has no arch')
        return ArchQuery.filename(name, self._epoch(), version, self.release(),
                                  arch)

    def gettag(self, tag):
        # implement me, if needed
        return None

    @staticmethod
    def query(filename, all_tags=False, *extra_tags):
        f = open(filename, 'rb')
        archq = ArchQuery(f)
        archq.read(all_tags, *extra_tags)
        f.close()
        return archq

    @staticmethod
    def rpmvercmp(ver1, ver2):
        """
        implementation of RPM's version comparison algorithm
        (as described in lib/rpmvercmp.c)
        """
        if ver1 == ver2:
            return 0
        elif ver1 is None:
            return -1
        elif ver2 is None:
            return 1
        res = 0
        while res == 0:
            # remove all leading non alphanumeric chars
            ver1 = re.sub(b'^[^a-zA-Z0-9]*', b'', ver1)
            ver2 = re.sub(b'^[^a-zA-Z0-9]*', b'', ver2)
            if not (len(ver1) and len(ver2)):
                break
            # check if we have a digits segment
            mo1 = re.match(br'(\d+)', ver1)
            mo2 = re.match(br'(\d+)', ver2)
            numeric = True
            if mo1 is None:
                mo1 = re.match(b'([a-zA-Z]+)', ver1)
                mo2 = re.match(b'([a-zA-Z]+)', ver2)
                numeric = False
            # check for different types: alpha and numeric
            if mo2 is None:
                if numeric:
                    return 1
                return -1
            seg1 = mo1.group(0)
            ver1 = ver1[mo1.end(0):]
            seg2 = mo2.group(1)
            ver2 = ver2[mo2.end(1):]
            if numeric:
                # remove leading zeros
                seg1 = re.sub(b'^0+', b'', seg1)
                seg2 = re.sub(b'^0+', b'', seg2)
                # longer digit segment wins - if both have the same length
                # a simple ascii compare decides
                res = len(seg1) - len(seg2) or packagequery.cmp(seg1, seg2)
            else:
                res = packagequery.cmp(seg1, seg2)
        if res > 0:
            return 1
        elif res < 0:
            return -1
        return packagequery.cmp(ver1, ver2)

    @staticmethod
    def filename(name, epoch, version, release, arch):
        if epoch:
            if release:
                return b'%s-%s:%s-%s-%s.arch' % (name, epoch, version, release, arch)
            else:
                return b'%s-%s:%s-%s.arch' % (name, epoch, version, arch)
        if release:
            return b'%s-%s-%s-%s.arch' % (name, version, release, arch)
        else:
            return b'%s-%s-%s.arch' % (name, version, arch)


if __name__ == '__main__':
    archq = ArchQuery.query(sys.argv[1])
    print(archq.name(), archq.version(), archq.release(), archq.arch())
    try:
        print(archq.canonname())
    except ArchError as e:
        print(e.msg)
    print(archq.description())
    print('##########')
    print(b'\n'.join(archq.provides()))
    print('##########')
    print(b'\n'.join(archq.requires()))



import os
import sys
import unittest

import osc.core
import osc.oscerr

from .common import OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'addfile_fixtures')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestAddFiles)


class TestAddFiles(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def testSimpleAdd(self):
        """add one file ('toadd1') to the wc"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.addfile('toadd1')
        exp = 'A    toadd1\n'
        self.assertEqual(sys.stdout.getvalue(), exp)
        self.assertFalse(os.path.exists(os.path.join('.osc', 'sources', 'toadd1')))
        self._check_status(p, 'toadd1', 'A')
        self._check_addlist('toadd1\n')

    def testSimpleMultipleAdd(self):
        """add multiple files ('toadd1', 'toadd2') to the wc"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.addfile('toadd1')
        p.addfile('toadd2')
        exp = 'A    toadd1\nA    toadd2\n'
        self.assertEqual(sys.stdout.getvalue(), exp)
        self.assertFalse(os.path.exists(os.path.join('.osc', 'sources', 'toadd1')))
        self.assertFalse(os.path.exists(os.path.join('.osc', 'sources', 'toadd2')))
        self._check_status(p, 'toadd1', 'A')
        self._check_status(p, 'toadd2', 'A')
        self._check_addlist('toadd1\ntoadd2\n')

    def testAddVersionedFile(self):
        """add a versioned file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        self.assertRaises(osc.oscerr.PackageFileConflict, p.addfile, 'merge')
        self.assertFalse(os.path.exists(os.path.join('.osc', '_to_be_added')))
        self._check_status(p, 'merge', ' ')

    def testAddUnversionedFileTwice(self):
        """add the same file twice"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.addfile('toadd1')
        self.assertRaises(osc.oscerr.PackageFileConflict, p.addfile, 'toadd1')
        exp = 'A    toadd1\n'
        self.assertEqual(sys.stdout.getvalue(), exp)
        self.assertFalse(os.path.exists(os.path.join('.osc', 'sources', 'toadd1')))
        self._check_status(p, 'toadd1', 'A')
        self._check_addlist('toadd1\n')

    def testReplace(self):
        """replace a deleted file ('foo')"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        with open('foo', 'w') as f:
            f.write('replaced file\n')
        p.addfile('foo')
        exp = 'A    foo\n'
        self.assertEqual(sys.stdout.getvalue(), exp)
        self.assertFileContentNotEqual(os.path.join('.osc', 'sources', 'foo'), 'replaced file\n')
        self.assertFalse(os.path.exists(os.path.join('.osc', '_to_be_deleted')))
        self._check_status(p, 'foo', 'R')
        self._check_addlist('foo\n')

    def testAddNonExistentFile(self):
        """add a non existent file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        self.assertRaises(osc.oscerr.OscIOError, p.addfile, 'doesnotexist')
        self.assertFalse(os.path.exists(os.path.join('.osc', '_to_be_added')))


if __name__ == '__main__':
    unittest.main()



import textwrap

from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .status_data import StatusData


class Status(XmlModel):
    XML_TAG = "status"

    code: str = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            Status code returned by the server.
            """
        ),
    )

    summary: Optional[str] = Field(
        description=textwrap.dedent(
            """
            Human readable summary.
            """
        ),
    )

    details: Optional[str] = Field(
        description=textwrap.dedent(
            """
            Detailed, human readable information.
            """
        ),
    )

    data_list: Optional[List[StatusData]] = Field(
        xml_name="data",
        description=textwrap.dedent(
            """
            Additional machine readable data.
            """
        ),
    )

    @property
    def data(self):
        result = {}
        for entry in self.data_list or []:
            key = entry.name
            value = entry.value
            result[key] = value
        return result



import os
import unittest

import osc.core
import osc.oscerr

from .common import OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'init_package_fixtures')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestInitPackage)


class TestInitPackage(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def setUp(self):
        super().setUp(copytree=False)

    def test_simple(self):
        """initialize a package dir"""
        pac_dir = os.path.join(self.tmpdir, 'testpkg')
        osc.core.Package.init_package('http://localhost', 'osctest', 'testpkg', pac_dir)
        storedir = os.path.join(pac_dir, osc.core.store)
        self.assertFalse(os.path.exists(os.path.join(storedir, '_meta_mode')))
        self.assertFalse(os.path.exists(os.path.join(storedir, '_size_limit')))
        self._check_list(os.path.join(storedir, '_project'), 'osctest\n')
        self._check_list(os.path.join(storedir, '_package'), 'testpkg\n')
        self._check_list(os.path.join(storedir, '_files'), '<directory />\n')
        self._check_list(os.path.join(storedir, '_apiurl'), 'http://localhost\n')

    def test_size_limit(self):
        """initialize a package dir with size_limit parameter"""
        pac_dir = os.path.join(self.tmpdir, 'testpkg')
        osc.core.Package.init_package('http://localhost', 'osctest', 'testpkg', pac_dir, size_limit=42)
        storedir = os.path.join(pac_dir, osc.core.store)
        self.assertFalse(os.path.exists(os.path.join(storedir, '_meta_mode')))
        self._check_list(os.path.join(storedir, '_size_limit'), '42\n')
        self._check_list(os.path.join(storedir, '_project'), 'osctest\n')
        self._check_list(os.path.join(storedir, '_package'), 'testpkg\n')
        self._check_list(os.path.join(storedir, '_files'), '<directory />\n')
        self._check_list(os.path.join(storedir, '_apiurl'), 'http://localhost\n')

    def test_meta_mode(self):
        """initialize a package dir with meta paramter"""
        pac_dir = os.path.join(self.tmpdir, 'testpkg')
        osc.core.Package.init_package('http://localhost', 'osctest', 'testpkg', pac_dir, meta=True)
        storedir = os.path.join(pac_dir, osc.core.store)
        self.assertFalse(os.path.exists(os.path.join(storedir, '_size_limit')))
        self._check_list(os.path.join(storedir, '_meta_mode'), '\n')
        self._check_list(os.path.join(storedir, '_project'), 'osctest\n')
        self._check_list(os.path.join(storedir, '_package'), 'testpkg\n')
        self._check_list(os.path.join(storedir, '_files'), '<directory />\n')
        self._check_list(os.path.join(storedir, '_apiurl'), 'http://localhost\n')

    def test_dirExists(self):
        """initialize a package dir (dir already exists)"""
        pac_dir = os.path.join(self.tmpdir, 'testpkg')
        os.mkdir(pac_dir)
        osc.core.Package.init_package('http://localhost', 'osctest', 'testpkg', pac_dir)
        storedir = os.path.join(pac_dir, osc.core.store)
        self.assertFalse(os.path.exists(os.path.join(storedir, '_meta_mode')))
        self.assertFalse(os.path.exists(os.path.join(storedir, '_size_limit')))
        self._check_list(os.path.join(storedir, '_project'), 'osctest\n')
        self._check_list(os.path.join(storedir, '_package'), 'testpkg\n')
        self._check_list(os.path.join(storedir, '_files'), '<directory />\n')
        self._check_list(os.path.join(storedir, '_apiurl'), 'http://localhost\n')

    def test_storedirExists(self):
        """initialize a package dir (dir+storedir already exists)"""
        pac_dir = os.path.join(self.tmpdir, 'testpkg')
        os.mkdir(pac_dir)
        os.mkdir(os.path.join(pac_dir, osc.core.store))
        self.assertRaises(osc.oscerr.OscIOError, osc.core.Package.init_package, 'http://localhost', 'osctest', 'testpkg', pac_dir)

    def test_dirIsFile(self):
        """initialize a package dir (dir is a file)"""
        pac_dir = os.path.join(self.tmpdir, 'testpkg')
        os.mkdir(pac_dir)
        with open(os.path.join(pac_dir, osc.core.store), 'w') as f:
            f.write('foo\n')
        self.assertRaises(osc.oscerr.OscIOError, osc.core.Package.init_package, 'http://localhost', 'osctest', 'testpkg', pac_dir)


if __name__ == '__main__':
    unittest.main()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import

from .flag import Flag
from .group_role import GroupRole
from .package_devel import PackageDevel
from .package_revision import PackageRevision
from .person_role import PersonRole
from .simple_flag import SimpleFlag
from .status import Status


class Package(XmlModel):
    XML_TAG = "package"

    name: str = Field(
        xml_attribute=True,
    )

    project: str = Field(
        xml_attribute=True,
    )

    title: Optional[str] = Field()

    description: Optional[str] = Field()

    devel: Optional[PackageDevel] = Field()

    releasename: Optional[str] = Field()

    person_list: Optional[List[PersonRole]] = Field(
        xml_name="person",
    )

    group_list: Optional[List[GroupRole]] = Field(
        xml_name="group",
    )

    lock: Optional[SimpleFlag] = Field(
        xml_wrapped=True,
    )

    build_list: Optional[List[Flag]] = Field(
        xml_name="build",
        xml_wrapped=True,
    )

    publish_list: Optional[List[Flag]] = Field(
        xml_name="publish",
        xml_wrapped=True,
    )

    useforbuild_list: Optional[List[Flag]] = Field(
        xml_name="useforbuild",
        xml_wrapped=True,
    )

    debuginfo_list: Optional[List[Flag]] = Field(
        xml_name="debuginfo",
        xml_wrapped=True,
    )

    binarydownload: Optional[SimpleFlag] = Field()

    sourceaccess: Optional[SimpleFlag] = Field()

    url: Optional[str] = Field()

    scmsync: Optional[str] = Field()

    bcntsynctag: Optional[str] = Field()

    @classmethod
    def from_api(cls, apiurl, project, package, *, rev=None):
        # ``rev`` is metadata revision, not revision of the source code
        url_path = ["source", project, package, "_meta"]
        url_query = {
            "rev": rev,
        }
        response = cls.xml_request("GET", apiurl, url_path, url_query)
        return cls.from_file(response, apiurl=apiurl)

    def to_api(self, apiurl, *, project=None, package=None):
        project = project or self.project
        package = package or self.name
        url_path = ["source", project, package, "_meta"]
        url_query = {}
        response = self.xml_request("PUT", apiurl, url_path, url_query, data=self.to_string())
        return Status.from_file(response, apiurl=apiurl)

    @classmethod
    def cmd_fork(
        cls,
        apiurl: str,
        project: str,
        package: str,
        *,
        scmsync: str,
    ):
        """
        POST /source/{project}/{package}?cmd=fork&scmsync={scmsync}
        For a package managed in Git.

        :param apiurl: Full apiurl or its alias.
        :param project: Project name.
        :param package: Package name.
        :param scmsync: Checkout Git URL. Example: https://src.example.com/owner/repo#branch
        """

        url_path = ["source", project, package]
        url_query = {
            "cmd": "fork",
            "scmsync": scmsync,
        }
        response = cls.xml_request("POST", apiurl, url_path, url_query)
        return Status.from_file(response, apiurl=apiurl)

    @classmethod
    def cmd_release(
        cls,
        apiurl: str,
        project: str,
        package: str,
        *,
        repository: Optional[str] = None,
        arch: Optional[str] = None,
        target_project: Optional[str] = None,
        target_repository: Optional[str] = None,
        setrelease: Optional[str] = None,
        nodelay: Optional[bool] = None,
    ):
        """
        POST /source/{project}/{package}?cmd=release
        Release sources and binaries of a specified package.

        :param apiurl: Full apiurl or its alias.
        :param project: Project name.
        :param package: Package name.
        :param repository: Limit the release to the given repository.
        :param arch: Limit the release to the given architecture.
        :param target_project: The name of the release target project.
        :param target_repository: The name of the release target repository.
        :param setrelease: Tag the release with the given value.
        :param nodelay: Do not delay the relase. If not set, the release will be delayed to be done later.
        """

        url_path = ["source", project, package]
        url_query = {
            "cmd": "release",
            "repository": repository,
            "arch": arch,
            "target_project": target_project,
            "target_repository": target_repository,
            "setrelease": setrelease,
            "nodelay": nodelay,
        }
        response = cls.xml_request("POST", apiurl, url_path, url_query)
        return Status.from_file(response, apiurl=apiurl)

    @classmethod
    def get_revision_list(cls, apiurl: str, project: str, package: str, deleted: Optional[bool] = None, meta: Optional[bool] = None):
        from ..util.xml import xml_parse

        url_path = ["source", project, package, "_history"]
        url_query = {
            "meta": meta,
            "deleted": deleted,
        }
        response = cls.xml_request("GET", apiurl, url_path, url_query)
        root = xml_parse(response).getroot()
        assert root.tag == "revisionlist"
        result = []
        for node in root:
            result.append(PackageRevision.from_xml(node, apiurl=apiurl))
        return result



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .enums import ObsRatings
from .request_action import RequestAction
from .request_history import RequestHistory
from .request_review import RequestReview
from .request_state import RequestState


class Request(XmlModel):
    XML_TAG = "request"

    id: Optional[str] = Field(
        xml_attribute=True,
    )

    actions: Optional[int] = Field(
        xml_attribute=True,
    )

    creator: Optional[str] = Field(
        xml_attribute=True,
    )

    action_list: List[RequestAction] = Field(
        xml_name="action",
    )

    state: Optional[RequestState] = Field(
    )

    description: Optional[str] = Field(
    )

    priority: Optional[ObsRatings] = Field(
    )

    review_list: Optional[List[RequestReview]] = Field(
        xml_name="review",
    )

    history_list: Optional[List[RequestHistory]] = Field(
        xml_name="history",
    )

    title: Optional[str] = Field(
    )

    accept_at: Optional[str] = Field(
    )

    @classmethod
    def from_api(
        cls,
        apiurl: str,
        request_id: int,
        *,
        with_history: Optional[bool] = None,
        with_full_history: Optional[bool] = None
    ) -> "Request":
        """
        Return the specified request.

        :param request_id: Id of the request.
        :param withhistory: Include the request history in the results.
        :param withfullhistory: Includes both, request and review history in the results.
        """
        url_path = ["request", request_id]
        url_query = {
            "withhistory": with_history,
            "withfullhistory": with_full_history,
        }
        response = cls.xml_request("GET", apiurl, url_path, url_query)
        return cls.from_file(response, apiurl=apiurl)

    @classmethod
    def cmd_diff(
        cls,
        apiurl: str,
        request_id: int,
        *,
        with_issues: Optional[bool] = None,
        with_description_issues: Optional[bool] = None,
        diff_to_superseded: Optional[int] = None
    ) -> "Request":
        """
        Return the specified request including a diff of all packages in the request.

        :param request_id: Id of the request.
        :param with_issues: Include parsed issues from referenced sources in the change files.
        :param with_description_issues: Include parsed issues from request description.
        :param diff_to_superseded: Diff relatively to the given superseded request.
        """
        url_path = ["request", str(request_id)]
        url_query = {
            "cmd": "diff",
            "view": "xml",
            "withissues": with_issues,
            "withdescriptionissues": with_description_issues,
            "diff_to_superseded": diff_to_superseded,
        }
        response = cls.xml_request("POST", apiurl, url_path, url_query)
        return cls.from_file(response, apiurl=apiurl)

    def get_issues(self):
        """
        Aggregate issues from action/sourcediff into a single list.
        The list may contain duplicates.

        To get any issues returned, it is crucial to load the request with the issues
        by calling ``cmd_diff()`` with appropriate arguments first.
        """
        result = []
        for action in self.action_list or []:
            if action.sourcediff is None:
                continue
            for issue in action.sourcediff.issue_list or []:
                result.append(issue)
        return result

    def cmd_create(self,
        apiurl: str,
        *,
        add_revision: Optional[bool] = None,
        enforce_branching: Optional[bool] = None,
        ignore_build_state: Optional[bool] = None,
        ignore_delegate: Optional[bool] = None,
    ):
        """
        :param add_revision: Ask the server to add revisions of the current sources to the request.
        :param ignore_build_state: Skip the build state check.
        :param ignore_delegate: Enforce a new package instance in a project which has OBS:DelegateRequestTarget set.
        """
        url_path = ["request"]
        url_query = {
            "cmd": "create",
            "addrevision": add_revision,
            "ignore_delegate": ignore_delegate,
        }
        response = self.xml_request("POST", apiurl, url_path, url_query, data=self.to_string())
        return Request.from_file(response, apiurl=apiurl)



import re
from typing import List
from typing import Optional
from typing import Tuple

from .connection import Connection
from .connection import GiteaHTTPResponse


class PullRequest:
    @classmethod
    def cmp(cls, entry: dict):
        if "base" in entry:
            # a proper pull request
            return entry["base"]["repo"]["full_name"], entry["number"]
        else:
            # an issue without pull request details
            return entry["repository"]["full_name"], entry["number"]

    @classmethod
    def split_id(cls, pr_id: str) -> Tuple[str, str, str]:
        """
        Split <owner>/<repo>#<number> into individual components and return them in a tuple.
        """
        match = re.match(r"^([^/]+)/([^/]+)#([0-9]+)$", pr_id)
        if not match:
            raise ValueError(f"Invalid pull request id: {pr_id}")
        return match.groups()

    @classmethod
    def to_human_readable_string(cls, entry: dict):
        from osc.output import KeyValueTable
        from . import User

        def yes_no(value):
            return "yes" if value else "no"

        if "base" in entry:
            # a proper pull request
            entry_id = f"{entry['base']['repo']['full_name']}#{entry['number']}"
            is_pull_request = True
        else:
            # an issue without pull request details
            entry_id = f"{entry['repository']['full_name']}#{entry['number']}"
            is_pull_request = False

        # HACK: search API returns issues, the URL needs to be transformed to a pull request URL
        entry_url = entry["url"]
        entry_url = re.sub(r"^(.*)/api/v1/repos/(.+/.+)/issues/([0-9]+)$", r"\1/\2/pulls/\3", entry_url)

        table = KeyValueTable()
        table.add("ID", entry_id, color="bold")
        table.add("URL", f"{entry_url}")
        table.add("Title", f"{entry['title']}")
        table.add("State", entry["state"])
        if is_pull_request:
            table.add("Draft", yes_no(entry["draft"]))
            table.add("Merged", yes_no(entry["merged"]))
            table.add("Allow edit", yes_no(entry["allow_maintainer_edit"]))
        table.add("Author", f"{User.to_login_full_name_email_string(entry['user'])}")
        if is_pull_request:
            table.add("Source", f"{entry['head']['repo']['full_name']}, branch: {entry['head']['ref']}, commit: {entry['head']['sha']}")
        table.add("Description", entry["body"])

        return str(table)

    @classmethod
    def list_to_human_readable_string(cls, entries: List, sort: bool = False):
        if sort:
            entries = sorted(entries, key=cls.cmp)
        result = []
        for entry in entries:
            result.append(cls.to_human_readable_string(entry))
        return "\n\n".join(result)

    @classmethod
    def create(
        cls,
        conn: Connection,
        *,
        target_owner: str,
        target_repo: str,
        target_branch: str,
        source_owner: str,
        source_branch: str,
        title: str,
        description: Optional[str] = None,
    ) -> GiteaHTTPResponse:
        """
        Create a pull request to ``owner``/``repo`` to the ``base`` branch.
        The pull request comes from a fork. The fork repo name is determined from gitea database.

        :param conn: Gitea ``Connection`` instance.
        :param target_owner: Owner of the target repo.
        :param target_repo: Name of the target repo.
        :param target_branch: Name of the target branch in the target repo.
        :param source_owner: Owner of the source (forked) repo.
        :param source_branch: Name of the source branch in the source (forked) repo.
        :param title: Pull request title.
        :param description: Pull request description.
        """
        url = conn.makeurl("repos", target_owner, target_repo, "pulls")
        data = {
            "base": target_branch,
            "head": f"{source_owner}:{source_branch}",
            "title": title,
            "body": description,
        }
        return conn.request("POST", url, json_data=data)

    @classmethod
    def get(
        cls,
        conn: Connection,
        owner: str,
        repo: str,
        number: int,
    ) -> GiteaHTTPResponse:
        """
        Get a pull request.

        :param conn: Gitea ``Connection`` instance.
        :param owner: Owner of the repo.
        :param repo: Name of the repo.
        :param number: Number of the pull request in the repo.
        """
        url = conn.makeurl("repos", owner, repo, "pulls", str(number))
        return conn.request("GET", url)

    @classmethod
    def set(
        cls,
        conn: Connection,
        owner: str,
        repo: str,
        number: int,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
        allow_maintainer_edit: Optional[bool] = None,
    ) -> GiteaHTTPResponse:
        """
        Change a pull request.

        :param conn: Gitea ``Connection`` instance.
        :param owner: Owner of the repo.
        :param repo: Name of the repo.
        :param number: Number of the pull request in the repo.
        :param title: Change pull request title.
        :param description: Change pull request description.
        :param allow_maintainer_edit: Change whether users with write access to the base branch can also push to the pull request's head branch.
        """
        json_data = {
            "title": title,
            "description": description,
            "allow_maintainer_edit": allow_maintainer_edit,
        }
        url = conn.makeurl("repos", owner, repo, "pulls", str(number))
        return conn.request("PATCH", url, json_data=json_data)

    @classmethod
    def list(
        cls,
        conn: Connection,
        owner: str,
        repo: str,
        *,
        state: Optional[str] = "open",
    ) -> GiteaHTTPResponse:
        """
        List pull requests in a repo.

        :param conn: Gitea ``Connection`` instance.
        :param owner: Owner of the repo.
        :param repo: Name of the repo.
        :param state: Filter by state: open, closed, all. Defaults to open.
        """
        if state == "all":
            state = None

        q = {
            "state": state,
        }
        url = conn.makeurl("repos", owner, repo, "pulls", query=q)
        return conn.request("GET", url)

    @classmethod
    def search(
        cls,
        conn: Connection,
        *,
        state: str = "open",
        title: Optional[str] = None,
        owner: Optional[str] = None,
        labels: Optional[List[str]] = None,
        assigned: bool = False,
        created: bool = False,
        mentioned: bool = False,
        review_requested: bool = False,
    ) -> GiteaHTTPResponse:
        """
        Search pull requests.
        :param conn: Gitea ``Connection`` instance.
        :param state: Filter by state: open, closed. Defaults to open.
        :param title: Filter by substring in title.
        :param owner: Filter by owner of the repository associated with the pull requests.
        :param labels: Filter by associated labels. Non existent labels are discarded.
        :param assigned: Filter pull requests assigned to you.
        :param created: Filter pull requests created by you.
        :param mentioned: Filter pull requests mentioning you.
        :param review_requested: Filter pull requests requesting your review.
        """
        q = {
            "type": "pulls",
            "state": state,
            "q": title,
            "owner": owner,
            "labels": ",".join(labels) if labels else None,
            "assigned": assigned,
            "created": created,
            "mentioned": mentioned,
            "review_requested": review_requested,
        }
        url = conn.makeurl("repos", "issues", "search", query=q)
        return conn.request("GET", url)

    @classmethod
    def get_patch(
        cls,
        conn: Connection,
        owner: str,
        repo: str,
        number: str,
    ) -> GiteaHTTPResponse:
        """
        Get a patch associated with a pull request.

        :param conn: Gitea ``Connection`` instance.
        :param owner: Owner of the repo.
        :param repo: Name of the repo.
        :param number: Number of the pull request in the repo.
        """
        url = conn.makeurl("repos", owner, repo, "pulls", f"{number}.patch")
        return conn.request("GET", url)



from .keyinfo import Keyinfo
from .package import Package
from .package_sources import PackageSources
from .person import Person
from .project import Project
from .request import Request
from .token import Token



import sys

import osc.commandline_git


class PullRequestGetCommand(osc.commandline_git.GitObsCommand):
    """
    Get details about the specified pull requests
    """

    name = "get"
    aliases = ["show"]  # for compatibility with osc
    parent = "PullRequestCommand"

    def init_arguments(self):
        self.add_argument_owner_repo_pull(nargs="+")
        self.add_argument(
            "-p",
            "--patch",
            action="store_true",
            help="Show patches associated with the pull requests",
        )

    def run(self, args):
        from osc import gitea_api
        from osc.core import highlight_diff
        from osc.output import tty

        self.print_gitea_settings()

        num_entries = 0
        failed_entries = []
        for owner, repo, pull in args.owner_repo_pull:
            try:
                pr = gitea_api.PullRequest.get(self.gitea_conn, owner, repo, int(pull)).json()
                num_entries += 1
            except gitea_api.GiteaException as e:
                if e.status == 404:
                    failed_entries.append(f"{owner}/{repo}#{pull}")
                    continue
                raise
            print(gitea_api.PullRequest.to_human_readable_string(pr))

            if args.patch:
                print("")
                print(tty.colorize("Patch:", "bold"))
                patch = gitea_api.PullRequest.get_patch(self.gitea_conn, owner, repo, pull).data
                patch = highlight_diff(patch)
                print(patch.decode("utf-8"))

            print()

        print(f"Total entries: {num_entries}", file=sys.stderr)
        if failed_entries:
            print(
                f"{tty.colorize('ERROR', 'red,bold')}: Couldn't retrieve the following pull requests: {', '.join(failed_entries)}",
                file=sys.stderr,
            )
            sys.exit(1)



from . import api
from .common import format_msg_project_package_options
from .. import oscerr
from ..output import print_msg


def add_channels(apiurl, project, package=None, enable_all=False, skip_disabled=False, print_to="debug"):
    if all((enable_all, skip_disabled)):
        raise oscerr.OscValueError("Options 'enable_all' and 'skip_disabled' are mutually exclusive")

    msg = format_msg_project_package_options(
        "Adding channels to",
        project,
        package,
        enable_all=enable_all,
        skip_disabled=skip_disabled,
    )
    print_msg(msg, print_to=print_to)

    url_path = ["source", project]
    if package:
        url_path += [package]

    url_query = {"cmd": "addchannels"}
    if enable_all:
        url_query["mode"] = "enable_all"
    if skip_disabled:
        url_query["mode"] = "skip_disabled"

    return api.post(apiurl, url_path, url_query)


def add_containers(apiurl, project, package, extend_package_names=False, print_to="debug"):
    msg = format_msg_project_package_options(
        "Adding containers to",
        project,
        package,
        extend_package_names=extend_package_names,
    )
    print_msg(msg, print_to=print_to)

    url_path = ["source", project, package]

    url_query = {"cmd": "addcontainers"}
    if extend_package_names:
        url_query["extend_package_names"] = "1"

    return api.post(apiurl, url_path, url_query)


def enable_channels(apiurl, project, package=None, print_to="debug"):
    msg = format_msg_project_package_options(
        "Enabling channels in",
        project,
        package,
    )
    print_msg(msg, print_to=print_to)

    url_path = ["source", project]
    if package:
        url_path += [package]

    if package:
        url_query = {"cmd": "enablechannel"}
    else:
        url_query = {"cmd": "modifychannels", "mode": "enable_all"}

    return api.post(apiurl, url_path, url_query)


def get_linked_packages(apiurl, project, package):
    url_path = ["source", project, package]
    url_query = {"cmd": "showlinked"}
    root = api.post(apiurl, url_path, url_query)

    result = []
    nodes = api.find_nodes(root, "collection", "package")
    for node in nodes:
        item = {
            "project": node.get("project"),
            "name": node.get("name"),
        }
        result.append(item)
    return result


def release(
    apiurl,
    project,
    package,
    repository,
    architecture,
    target_project,
    target_repository,
    set_release_to=None,
    delayed=False,
    print_to="debug",
):
    msg = format_msg_project_package_options(
        "Releasing",
        project,
        package,
        target_project,
        target_package=None,
        repository=repository,
        architecture=architecture,
        dest_repository=target_repository,
        delayed=delayed,
    )
    print_msg(msg, print_to=print_to)

    url_path = ["source", project]
    if package:
        url_path += [package]

    url_query = {"cmd": "release"}
    if repository:
        url_query["repository"] = repository
    if architecture:
        url_query["arch"] = architecture
    if target_project:
        url_query["target_project"] = target_project
    if target_repository:
        url_query["target_repository"] = target_repository
    if set_release_to:
        url_query["setrelease"] = set_release_to
    if not delayed:
        url_query["nodelay"] = "1"

    return api.post(apiurl, url_path, url_query)



import osc.commandline


class PersonRegisterCommand(osc.commandline.OscCommand):
    """
    Register a new person (user)
    """

    name = "register"
    parent = "PersonCommand"

    def init_arguments(self):
        self.add_argument(
            "--login",
            required=True,
            help="Login.",
        )
        self.add_argument(
            "--realname",
            required=True,
            help="Real name of the person.",
        )
        self.add_argument(
            "--email",
            required=True,
            help="Email address.",
        )
        self.add_argument(
            "--password",
            help="Password. An interactive prompt is shown if password is not specified.",
        )
        self.add_argument(
            "--note",
            help="Any notes about the person.",
        )
        self.add_argument(
            "--state",
            help="State of the account. Defaults to 'unconfirmed'.",
        )

    def run(self, args):
        from osc import obs_api
        from osc.util.helper import raw_input

        if args.password:
            password = args.password
        else:
            password = raw_input(f"Enter password for {args.login}@{args.apiurl}: ")

        obs_api.Person.cmd_register(
            args.apiurl,
            login=args.login,
            realname=args.realname,
            email=args.email,
            password=password,
            note=args.note,
            state=args.state,
        )



import base64
import os
from tempfile import mkdtemp
from shutil import rmtree


class KeyError(Exception):
    def __init__(self, key, *args):
        super().__init__()
        self.args = args
        self.key = key

    def __str__(self):
        return f"{self.key} :{' '.join(self.args)}"


class Checker:
    def __init__(self):
        import rpm
        self.dbdir = mkdtemp(prefix='oscrpmdb')
        self.imported = {}
        # pylint: disable=E1101
        rpm.addMacro('_dbpath', self.dbdir)
        self.ts = rpm.TransactionSet()
        self.ts.initDB()
        self.ts.openDB()
        self.ts.setVSFlags(0)
        # self.ts.Debug(1)

    def readkeys(self, keys=None):
        import rpm
        keys = keys or []
        # pylint: disable=E1101
        rpm.addMacro('_dbpath', self.dbdir)
        for key in keys:
            try:
                self.readkey(key)
            except KeyError as e:
                print(e)

        if not self.imported:
            raise KeyError('', "no key imported")

        import rpm
        # pylint: disable=E1101
        rpm.delMacro("_dbpath")

# python is an idiot
#    def __del__(self):
#        self.cleanup()

    def cleanup(self):
        self.ts.closeDB()
        rmtree(self.dbdir)

    def readkey(self, file):
        if file in self.imported:
            return

        fd = open(file)
        line = fd.readline()
        if line and line[0:14] == "-----BEGIN PGP":
            line = fd.readline()
            while line and line != "\n":
                line = fd.readline()
            if not line:
                raise KeyError(file, "not a pgp public key")
        else:
            raise KeyError(file, "not a pgp public key")

        key = ''
        line = fd.readline()
        crc = None
        while line:
            if line[0:12] == "-----END PGP":
                break
            line = line.rstrip()
            if line[0] == '=':
                crc = line[1:]
                line = fd.readline()
                break
            else:
                key += line
                line = fd.readline()
        fd.close()
        if not line or line[0:12] != "-----END PGP":
            raise KeyError(file, "not a pgp public key")

        # TODO: compute and compare CRC, see RFC 2440

        bkey = base64.b64decode(key)

        r = self.ts.pgpImportPubkey(bkey)
        if r != 0:
            raise KeyError(file, "failed to import pubkey")
        self.imported[file] = 1

    def check(self, pkg):
        # avoid errors on non rpm
        if pkg[-4:] != '.rpm':
            return
        fd = None
        try:
            fd = os.open(pkg, os.O_RDONLY)
            hdr = self.ts.hdrFromFdno(fd)
        finally:
            if fd is not None:
                os.close(fd)


if __name__ == "__main__":
    import sys
    keyfiles = []
    pkgs = []
    for arg in sys.argv[1:]:
        if arg[-4:] == '.rpm':
            pkgs.append(arg)
        else:
            keyfiles.append(arg)

    checker = Checker()
    try:
        checker.readkeys(keyfiles)
        for pkg in pkgs:
            checker.check(pkg)
    except Exception as e:
        checker.cleanup()
        raise e

# vim: sw=4 et



import os
import shutil
import tempfile
import unittest

from osc.core import binary_file
from osc.core import makeurl
from osc.core import UrlQueryArray
from osc.core import parseRevisionOption
from osc.oscerr import OscInvalidRevision


class TestParseRevisionOption(unittest.TestCase):
    def test_empty(self):
        expected = None, None
        actual = parseRevisionOption("")
        self.assertEqual(expected, actual)

    def test_colon(self):
        expected = None, None
        actual = parseRevisionOption(":")
        # your revision ':' will be ignored
        self.assertEqual(expected, actual)

    def test_invalid_multiple_colons(self):
        self.assertRaises(OscInvalidRevision, parseRevisionOption, ":::::")

    def test_one_number(self):
        expected = ("1", None)
        actual = parseRevisionOption("1")
        self.assertEqual(expected, actual)

    def test_two_numbers(self):
        expected = ("1", "2")
        actual = parseRevisionOption("1:2")
        self.assertEqual(expected, actual)

    def test_invalid_multiple_numbers(self):
        self.assertRaises(OscInvalidRevision, parseRevisionOption, "1:2:3:4:5")

    def test_one_hash(self):
        expected = "c4ca4238a0b923820dcc509a6f75849b", None
        actual = parseRevisionOption("c4ca4238a0b923820dcc509a6f75849b")
        self.assertEqual(expected, actual)

    def test_two_hashes(self):
        expected = ("d41d8cd98f00b204e9800998ecf8427e", "c4ca4238a0b923820dcc509a6f75849b")
        actual = parseRevisionOption("d41d8cd98f00b204e9800998ecf8427e:c4ca4238a0b923820dcc509a6f75849b")
        self.assertEqual(expected, actual)

    def test_invalid_multiple_hashes(self):
        rev = "d41d8cd98f00b204e9800998ecf8427e:c4ca4238a0b923820dcc509a6f75849b:c81e728d9d4c2f636f067f89cc14862c"
        self.assertRaises(OscInvalidRevision, parseRevisionOption, rev)


class TestMakeurl(unittest.TestCase):
    def test_basic(self):
        url = makeurl("https://example.com/api/v1", ["path", "to", "resource"], {"k1": "v1", "k2": ["v2", "v3"]})
        self.assertEqual(url, "https://example.com/api/v1/path/to/resource?k1=v1&k2=v2&k2=v3")

    def test_array(self):
        url = makeurl("https://example.com/api/v1", ["path", "to", "resource"], {"k1": "v1", "k2": UrlQueryArray(["v2", "v3"])})
        self.assertEqual(url, "https://example.com/api/v1/path/to/resource?k1=v1&k2%5B%5D=v2&k2%5B%5D=v3")

    def test_query_none(self):
        url = makeurl("https://example.com/api/v1", [], {"none": None})
        self.assertEqual(url, "https://example.com/api/v1")

    def test_query_empty_list(self):
        url = makeurl("https://example.com/api/v1", [], {"empty_list": []})
        self.assertEqual(url, "https://example.com/api/v1")

    def test_query_int(self):
        url = makeurl("https://example.com/api/v1", [], {"int": 1})
        self.assertEqual(url, "https://example.com/api/v1?int=1")

    def test_query_bool(self):
        url = makeurl("https://example.com/api/v1", [], {"bool": True})
        self.assertEqual(url, "https://example.com/api/v1?bool=1")

        url = makeurl("https://example.com/api/v1", [], {"bool": False})
        self.assertEqual(url, "https://example.com/api/v1?bool=0")

    def test_quote_path(self):
        mapping = (
            # (character, expected encoded character)
            (" ", "%20"),
            ("!", "%21"),
            ('"', "%22"),
            ("#", "%23"),
            ("$", "%24"),
            ("%", "%25"),
            ("&", "%26"),
            ("'", "%27"),
            ("(", "%28"),
            (")", "%29"),
            ("*", "%2A"),
            ("+", "%2B"),
            (",", "%2C"),
            ("/", "/"),
            (":", ":"),  # %3A
            (";", "%3B"),
            ("=", "%3D"),
            ("?", "%3F"),
            ("@", "%40"),
            ("[", "%5B"),
            ("]", "%5D"),
        )

        for char, encoded_char in mapping:
            url = makeurl("https://example.com/api/v1", [f"PREFIX_{char}_SUFFIX"])
            self.assertEqual(url, f"https://example.com/api/v1/PREFIX_{encoded_char}_SUFFIX")

    def test_quote_query(self):
        mapping = (
            # (character, expected encoded character)
            (" ", "+"),
            ("!", "%21"),
            ('"', "%22"),
            ("#", "%23"),
            ("$", "%24"),
            ("%", "%25"),
            ("&", "%26"),
            ("'", "%27"),
            ("(", "%28"),
            (")", "%29"),
            ("*", "%2A"),
            ("+", "%2B"),
            (",", "%2C"),
            ("/", "%2F"),
            (":", "%3A"),
            (";", "%3B"),
            ("=", "%3D"),
            ("?", "%3F"),
            ("@", "%40"),
            ("[", "%5B"),
            ("]", "%5D"),
        )

        for char, encoded_char in mapping:
            url = makeurl("https://example.com/api/v1", [], {char: char})
            self.assertEqual(url, f"https://example.com/api/v1?{encoded_char}={encoded_char}")


class TestBinaryFile(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix="osc_test_")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_text(self):
        path = os.path.join(self.tmpdir, "text")
        with open(path, "w") as f:
            f.write(1000 * "a")
        self.assertFalse(binary_file(path))

    def test_text_with_binary_chars(self):
        path = os.path.join(self.tmpdir, "binary")
        with open(path, "wb") as f:
            f.write(1000 * b"a")
            f.write(b"\1")
        self.assertFalse(binary_file(path))

        with open(path, "wb") as f:
            f.write(4096 * b"a")
            f.write(b"\1")
        self.assertFalse(binary_file(path))

    def test_binary(self):
        path = os.path.join(self.tmpdir, "binary")

        # sufficient control chars in first 4k
        with open(path, "wb") as f:
            f.write(1000 * b"a")
            f.write(26 * b"\1")
        self.assertTrue(binary_file(path))

        # sufficient control chars in first 4k
        with open(path, "wb") as f:
            f.write(3993 * b"a")
            f.write(103 * b"\1")
        self.assertTrue(binary_file(path))

        # a single \0 is good enough for us to say it's a binary file
        with open(path, "wb") as f:
            f.write(3993 * b"a")
            f.write(b"\0")
            f.write(999 * b"\1")
        self.assertTrue(binary_file(path))

        # detected as text because we're reading only first 4k characters
        with open(path, "wb") as f:
            f.write(4096 * b"a")
            f.write(1000 * b"\1")
        self.assertFalse(binary_file(path))


if __name__ == "__main__":
    unittest.main()



import typing

from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class ScmsyncObsinfo(BaseModel):
    """
    Class for handling _scmsync.obsinfo files
    """

    # the fields are defined in obs_scm_bridge in ObsGit.write_obsinfo()
    # https://github.com/openSUSE/obs-scm-bridge/blob/main/obs_scm_bridge
    mtime: int = Field()
    commit: str = Field()
    url: Optional[str] = Field()
    revision: Optional[str] = Field()
    subdir: Optional[str] = Field()
    projectscmsync: Optional[str] = Field()

    @classmethod
    def from_string(cls, data: str) -> "ScmsyncObsinfo":
        kwargs = {}
        for line in data.splitlines():
            line = line.strip()
            if not line:
                continue
            key, value = line.split(": ", 1)
            field = cls.__fields__.get(key, None)
            if field and field.type is int:
                value = int(value)
            kwargs[key] = value
        return cls(**kwargs)

    @classmethod
    def from_file(cls, file: Union[str, typing.IO]) -> "ScmsyncObsinfo":
        if isinstance(file, str):
            with open(file, "r", encoding="utf-8") as f:
                return cls.from_string(f.read())
        data = file.read()
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        return cls.from_string(data)

    @classmethod
    def from_api(cls, apiurl: str, project: str, package: str, *, rev: str) -> "ScmsyncObsinfo":
        import urllib.error
        from .. import oscerr
        from ..connection import http_request
        from ..core import makeurl

        url_path = ["source", project, package, "_scmsync.obsinfo"]
        url_query = {"rev": rev}
        url = makeurl(apiurl, url_path, url_query)
        try:
            response = http_request("GET", url)
        except urllib.error.HTTPError as e:
            if e.status == 404:
                raise oscerr.NotFoundAPIError(f"File '_scmsync.obsinfo' was not found in {project}/{package}, rev={rev}")
            raise
        return cls.from_file(response)



import io
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import ruamel.yaml

from osc import oscerr
from osc.util.models import BaseModel
from osc.util.models import Field


class Login(BaseModel):
    name: str = Field()  # type: ignore[assignment]
    url: str = Field()  # type: ignore[assignment]
    user: str = Field()  # type: ignore[assignment]
    token: str = Field()  # type: ignore[assignment]
    ssh_key: Optional[str] = Field()  # type: ignore[assignment]
    default: Optional[bool] = Field()  # type: ignore[assignment]

    class AlreadyExists(oscerr.OscBaseError):
        def __init__(self, name):
            super().__init__()
            self.name = name

        def __str__(self):
            return f"Gitea config entry with name '{self.name}' already exists"

    class DoesNotExist(oscerr.OscBaseError):
        def __init__(self, **kwargs):
            super().__init__()
            self.kwargs = kwargs

        def __str__(self):
            if self.kwargs == {"name": None}:
                return "Could not find a default Gitea config entry"
            kwargs_str = ", ".join([f"{key}={value}" for key, value in self.kwargs.items()])
            return f"Could not find a matching Gitea config entry: {kwargs_str}"

    def __init__(self, **kwargs):
        # ignore extra fields
        for key in list(kwargs):
            if key not in self.__fields__:
                kwargs.pop(key, None)
        super().__init__(**kwargs)

    def to_human_readable_string(self, *, show_token: bool = False):
        from osc.output import KeyValueTable

        table = KeyValueTable()
        table.add("Name", self.name, color="bold")
        if self.default:
            table.add("Default", "true", color="bold")
        table.add("URL", self.url)
        table.add("User", self.user)
        if self.ssh_key:
            table.add("SSH Key", self.ssh_key)
        if show_token:
            # tokens are stored in the plain text, there's not reason to protect them too much
            # let's only hide them from the output by default
            table.add("Token", self.token)
        return f"{table}"


class Config:
    """
    Manage the tea config.yml file.
    No data is cached in the objects, all changes are in sync with the file on disk.
    """

    def __init__(self, path: Optional[str] = None):
        if not path:
            path = os.path.expanduser("~/.config/tea/config.yml")
        self.path = os.path.abspath(path)

        self.logins: List[Login] = []

    def _read(self) -> Dict[str, Any]:
        try:
            with open(self.path, "r") as f:
                yaml = ruamel.yaml.YAML()
                return yaml.load(f)
        except FileNotFoundError:
            return {}

    def _write(self, data):
        yaml = ruamel.yaml.YAML()
        yaml.default_flow_style = False
        buf = io.StringIO()
        yaml.dump(data, buf)
        buf.seek(0)
        text = buf.read()

        os.makedirs(os.path.dirname(self.path), mode=0o700, exist_ok=True)
        with open(self.path, "w") as f:
            f.write(text)

    def list_logins(self) -> List[Login]:
        data = self._read()
        result = []
        for i in data.get("logins", []):
            login = Login(**i)
            result.append(login)
        return result

    def get_login(self, name: Optional[str] = None) -> Login:
        """
        Return ``Login`` object for the given ``name``.
        If ``name`` equals to ``None``, return the default ``Login``.
        """
        for login in self.list_logins():
            if name is None and login.default:
                return login
            if login.name == name:
                return login
        raise Login.DoesNotExist(name=name)

    def get_login_by_url_user(self, url: str, user: str) -> Login:
        """
        Return ``Login`` object for the given ``url`` and ``user``.
        """
        for login in self.list_logins():
            if (login.url, login.user) == (url, user):
                return login
        raise Login.DoesNotExist(url=url, user=user)

    def add_login(self, login: Login):
        data = self._read()
        data.setdefault("logins", [])

        for entry in data["logins"]:
            if entry.get("name", None) == login.name:
                raise Login.AlreadyExists(login.name)
            else:
                if login.default:
                    entry.pop("default", None)

        data["logins"].append(login.dict())
        self._write(data)

    def remove_login(self, name: str) -> Login:
        # throw an exception if the login name doesn't exist
        login = self.get_login(name)

        data = self._read()
        for num, entry in enumerate(list(data["logins"])):
            if entry.get("name", None) == login.name:
                data["logins"].pop(num)
        self._write(data)
        return login

    def update_login(
        self,
        name: str,
        new_name: Optional[str] = None,
        new_url: Optional[str] = None,
        new_user: Optional[str] = None,
        new_token: Optional[str] = None,
        new_ssh_key: Optional[str] = None,
        set_as_default: Optional[bool] = None,
    ) -> Login:
        login = self.get_login(name)

        if new_name is not None:
            login.name = new_name
        if new_url is not None:
            login.url = new_url
        if new_user is not None:
            login.user = new_user
        if new_token is not None:
            login.token = new_token
        if new_ssh_key is not None:
            login.ssh_key = new_ssh_key
        if set_as_default:
            login.default = True

        if not login.has_changed():
            return login

        data = self._read()
        for entry in data["logins"]:
            if entry.get("name", None) == name:
                entry.update(login.dict())
            else:
                if set_as_default:
                    entry.pop("default", None)
        self._write(data)

        return login



import os
import re
import subprocess
import sys
from typing import Optional

import osc.commandline_git


def get_editor() -> str:
    import shutil

    editor = os.getenv("EDITOR", None)
    if editor:
        candidates = [editor]
    else:
        candidates = ["vim", "vi"]

    editor_path = None
    for i in candidates:
        editor_path = shutil.which(i)
        if editor_path:
            break

    if not editor_path:
        raise RuntimeError(f"Unable to start editor '{candidates[0]}'")

    return editor_path


def run_editor(file_path: str):
    cmd = [get_editor(), file_path]
    subprocess.run(cmd)


def edit_message(template: Optional[str] = None) -> str:
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w+", encoding="utf-8", prefix="git_obs_message_") as f:
        if template:
            f.write(template)
            f.flush()

        run_editor(f.name)

        f.seek(0)
        return f.read()


NEW_PULL_REQUEST_TEMPLATE = """
# title
{title}

# description
{description}

#
# Please enter pull request title and description in the following format:
# <title>
# <blank line>
# <description>
#
# Lines starting with '#' will be ignored, and an empty message aborts the operation.
#
# Creating {source_owner}/{source_repo}#{source_branch} -> {target_owner}/{target_repo}#{target_branch}
#
""".lstrip()


class PullRequestCreateCommand(osc.commandline_git.GitObsCommand):
    """
    Create a pull request
    """

    name = "create"
    parent = "PullRequestCommand"

    def init_arguments(self):
        self.add_argument(
            "--title",
            metavar="TEXT",
            help="Pull request title",
        )
        self.add_argument(
            "--description",
            metavar="TEXT",
            help="Pull request description (body)",
        )
        self.add_argument(
            "--source-owner",
            metavar="OWNER",
            help="Owner of the source repo (default: derived from remote URL in local git repo)",
        )
        self.add_argument(
            "--source-repo",
            metavar="REPO",
            help="Name of the source repo (default: derived from remote URL in local git repo)",
        )
        self.add_argument(
            "--source-branch",
            metavar="BRANCH",
            help="Source branch (default: the current branch in local git repo)",
        )
        self.add_argument(
            "--target-branch",
            metavar="BRANCH",
            help="Target branch (default: derived from the current branch in local git repo)",
        )

    def run(self, args):
        from osc import gitea_api

        # the source args are optional, but if one of them is used, the others must be used too
        source_args = (args.source_owner, args.source_repo, args.source_branch)
        if sum((int(i is not None) for i in source_args)) not in (0, len(source_args)):
            self.parser.error("All of the following options must be used together: --source-owner, --source-repo, --source-branch")

        self.print_gitea_settings()

        use_local_git = args.source_owner is None

        if use_local_git:
            # local git repo
            git = gitea_api.Git(".")
            local_owner, local_repo = git.get_owner_repo()
            local_branch = git.current_branch
            local_rev = git.get_branch_head(local_branch)

        # remote git repo - source
        if use_local_git:
            source_owner = local_owner
            source_repo = local_repo
            source_branch = local_branch
        else:
            source_owner = args.source_owner
            source_repo = args.source_repo
            source_branch = args.source_branch
        source_repo_data = gitea_api.Repo.get(self.gitea_conn, source_owner, source_repo).json()
        source_branch_data = gitea_api.Branch.get(self.gitea_conn, source_owner, source_repo, source_branch).json()
        source_rev = source_branch_data["commit"]["id"]

        # remote git repo - target
        target_owner, target_repo = source_repo_data["parent"]["full_name"].split("/")

        if source_branch.startswith("for/"):
            # source branch name format: for/<target-branch>/<what-the-branch-name-would-normally-be>
            target_branch = source_branch.split("/")[1]
        else:
            target_branch = source_branch

        target_branch_data = gitea_api.Branch.get(self.gitea_conn, target_owner, target_repo, target_branch).json()
        target_rev = target_branch_data["commit"]["id"]

        print("Creating a pull request ...", file=sys.stderr)
        if use_local_git:
            print(f" * Local git: branch: {local_branch}, rev: {local_rev}", file=sys.stderr)
        print(f" * Source: {source_owner}/{source_repo}, branch: {source_branch}, rev: {source_rev}", file=sys.stderr)
        print(f" * Target: {target_owner}/{target_repo}, branch: {target_branch}, rev: {target_rev}", file=sys.stderr)

        if use_local_git and local_rev != source_rev:
            from osc.output import tty
            print(f"{tty.colorize('ERROR', 'red,bold')}: Local commit doesn't correspond with the latest commit in the remote source branch")
            sys.exit(1)

        if source_rev == target_rev:
            from osc.output import tty
            print(f"{tty.colorize('ERROR', 'red,bold')}: Source and target are identical, make and push changes to the remote source repo first")
            sys.exit(1)

        title = args.title or ""
        description = args.description or ""

        if not title or not description:
            # TODO: add list of commits and list of changed files to the template; requires local git repo
            message = edit_message(template=NEW_PULL_REQUEST_TEMPLATE.format(**locals()))

            # remove comments
            message = "\n".join([i for i in message.splitlines() if not i.startswith("#")])

            # strip leading and trailing spaces
            message = message.strip()

            if not message:
                raise RuntimeError("Aborting operation due to empty title and description.")

            parts = re.split(r"\n\n", message, 1)
            if len(parts) == 1:
                # empty description
                title = parts[0]
                description = ""
            else:
                title = parts[0]
                description = parts[1]

            title = title.strip()
            description = description.strip()

        pull = gitea_api.PullRequest.create(
            self.gitea_conn,
            target_owner=target_owner,
            target_repo=target_repo,
            target_branch=target_branch,
            source_owner=source_owner,
            # source_repo is not required because the information lives in Gitea database
            source_branch=source_branch,
            title=title,
            description=description,
        ).json()

        print("", file=sys.stderr)
        print("Pull request created:", file=sys.stderr)
        print(gitea_api.PullRequest.to_human_readable_string(pull))



import sys

import osc.commandline_git


class SSHKeyRemoveCommand(osc.commandline_git.GitObsCommand):
    """
    """

    name = "remove"
    parent = "SSHKeyCommand"

    def init_arguments(self):
        self.parser.add_argument(
            "id",
            type=int,
            help="Id of the SSH public key",
        )

    def run(self, args):
        from osc import gitea_api

        self.print_gitea_settings()

        print(f"Removing ssh key with id='{args.id}' ...", file=sys.stderr)
        response = gitea_api.SSHKey.get(self.gitea_conn, args.id)
        gitea_api.SSHKey.delete(self.gitea_conn, args.id)

        print("Removed entry:")
        print(gitea_api.SSHKey.to_human_readable_string(response.json()))



"""
Functions that communicate with OBS API
and work with related XML data.
"""


import xml.sax.saxutils
from xml.etree import ElementTree as ET

from ..util.xml import xml_escape
from ..util.xml import xml_indent
from ..util.xml import xml_unescape
from ..util.xml import xml_parse


def get(apiurl, path, query=None):
    """
    Send a GET request to OBS.

    :param apiurl: OBS apiurl.
    :type  apiurl: str
    :param path: URL path segments.
    :type  path: list(str)
    :param query: URL query values.
    :type  query: dict(str, str)
    :returns: Parsed XML root.
    :rtype:   xml.etree.ElementTree.Element
    """
    from .. import connection as osc_connection
    from .. import core as osc_core

    assert apiurl
    assert path

    if not isinstance(path, (list, tuple)):
        raise TypeError("Argument `path` expects a list of strings")

    url = osc_core.makeurl(apiurl, path, query)
    with osc_connection.http_GET(url) as f:
        root = xml_parse(f).getroot()
    return root


def post(apiurl, path, query=None):
    """
    Send a POST request to OBS.

    :param apiurl: OBS apiurl.
    :type  apiurl: str
    :param path: URL path segments.
    :type  path: list(str)
    :param query: URL query values.
    :type  query: dict(str, str)
    :returns: Parsed XML root.
    :rtype:   xml.etree.ElementTree.Element
    """
    from .. import connection as osc_connection
    from .. import core as osc_core

    assert apiurl
    assert path

    if not isinstance(path, (list, tuple)):
        raise TypeError("Argument `path` expects a list of strings")

    url = osc_core.makeurl(apiurl, path, query)
    with osc_connection.http_POST(url) as f:
        root = xml_parse(f).getroot()
    return root


def put(apiurl, path, query=None, data=None):
    """
    Send a PUT request to OBS.

    :param apiurl: OBS apiurl.
    :type  apiurl: str
    :param path: URL path segments.
    :type  path: list(str)
    :param query: URL query values.
    :type  query: dict(str, str)
    :returns: Parsed XML root.
    :rtype:   xml.etree.ElementTree.Element
    """
    from osc import connection as osc_connection
    from osc import core as osc_core

    assert apiurl
    assert path

    if not isinstance(path, (list, tuple)):
        raise TypeError("Argument `path` expects a list of strings")

    url = osc_core.makeurl(apiurl, path, query)
    with osc_connection.http_PUT(url, data=data) as f:
        root = xml_parse(f).getroot()
    return root


def _to_xpath(*args):
    """
    Convert strings and dictionaries to xpath:
        string gets translated to a node name
        dictionary gets translated to [@key='value'] predicate

    All values are properly escaped.

    Examples:
        args: ["directory", "entry", {"name": "osc"}]
        result: "directory/entry[@name='osc']"

        args: ["attributes", "attribute", {"namespace": "OBS", "name": "BranchSkipRepositories"}, "value"]
        result: "attributes/attribute[@namespace='OBS'][@name='BranchSkipRepositories']/value"
    """
    xpath = ""
    for arg in args:
        if isinstance(arg, str):
            arg = xml.sax.saxutils.escape(arg)
            xpath += f"/{arg}"
        elif isinstance(arg, dict):
            for key, value in arg.items():
                key = xml.sax.saxutils.escape(key)
                value = xml.sax.saxutils.escape(value)
                xpath += f"[@{key}='{value}']"
        else:
            raise TypeError(f"Argument '{arg}' has invalid type '{type(arg).__name__}'. Expected types: str, dict")

    # strip the leading slash because we're making a relative search
    xpath = xpath.lstrip("/")
    return xpath


def find_nodes(root, root_name, *args):
    """
    Find nodes with given `node_name`.
    Also, verify that the root tag matches the `root_name`.

    :param root: Root node.
    :type  root: xml.etree.ElementTree.Element
    :param root_name: Expected (tag) name of the root node.
    :type  root_name: str
    :param *args: Simplified xpath notation: strings are node names, dictionaries translate to [@key='value'] predicates.
    :type  *args: list[str, dict]
    :returns: List of nodes that match xpath based on the given `args`.
    :rtype:   list(xml.etree.ElementTree.Element)
    """
    assert root.tag == root_name
    return root.findall(_to_xpath(*args))


def find_node(root, root_name, *args):
    """
    Find a single node with given `node_name`.
    If `node_name` is not specified, the root node is returned.
    Also, verify that the root tag matches the `root_name`.

    :param root: Root node.
    :type  root: xml.etree.ElementTree.Element
    :param root_name: Expected (tag) name of the root node.
    :type  root_name: str
    :param *args: Simplified xpath notation: strings are node names, dictionaries translate to [@key='value'] predicates.
    :type  *args: list[str, dict]
    :returns: The node that matches xpath based on the given `args`
              or the root node if `args` are not specified.
    :rtype:   xml.etree.ElementTree.Element
    """

    assert root.tag == root_name
    if not args:
        # only verify the root tag
        return root
    return root.find(_to_xpath(*args))


def group_child_nodes(node):
    nodes = node[:]
    result = []

    while nodes:
        # look at the tag of the first node
        tag = nodes[0].tag

        # collect all nodes with the same tag and append them to the result
        # then repeat the step for the next tag(s)
        matches = []
        others = []
        for i in nodes:
            if i.tag == tag:
                matches.append(i)
            else:
                others.append(i)

        result += matches
        nodes = others

    node[:] = result


def write_xml_node_to_file(node, path, indent=True):
    """
    Write a XML node to a file.

    :param node: Node to write.
    :type  node: xml.etree.ElementTree.Element
    :param path: Path to a file that will be written to.
    :type  path: str
    :param indent: Whether to indent (pretty-print) the written XML.
    :type  indent: bool
    """
    if indent:
        xml_indent(node)
    ET.ElementTree(node).write(path)



from .connection import Connection
from .connection import GiteaHTTPResponse


class User:
    @classmethod
    def to_full_name_email_string(cls, data):
        full_name = data["full_name"]
        email = data["email"]
        if full_name:
            return f"{full_name} <{email}>"
        return email

    @classmethod
    def to_login_full_name_email_string(cls, data):
        return f"{data['login']} ({cls.to_full_name_email_string(data)})"

    @classmethod
    def get(
        cls,
        conn: Connection,
    ) -> GiteaHTTPResponse:
        """
        Retrieve details about the current user.

        :param conn: Gitea ``Connection`` instance.
        """
        url = conn.makeurl("user")
        return conn.request("GET", url)



import sys

import osc.commandline_git


class PullRequestListCommand(osc.commandline_git.GitObsCommand):
    """
    List pull requests in a repository
    """

    name = "list"
    parent = "PullRequestCommand"

    def init_arguments(self):
        self.add_argument_owner_repo(nargs="+")
        self.add_argument(
            "--state",
            choices=["open", "closed", "all"],
            default="open",
            help="State of the pull requests (default: open)",
        )

    def run(self, args):
        from osc import gitea_api

        self.print_gitea_settings()

        total_entries = 0
        for owner, repo in args.owner_repo:
            data = gitea_api.PullRequest.list(self.gitea_conn, owner, repo, state=args.state).json()
            total_entries += len(data)

            text = gitea_api.PullRequest.list_to_human_readable_string(data, sort=True)
            if text:
                print(text)
                print("", file=sys.stderr)

        print(f"Total entries: {total_entries}", file=sys.stderr)



import os
import unittest

from osc._private.package import ApiPackage
from osc._private.package import LocalPackage
from osc._private.package import PackageBase

from .common import GET
from .common import OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "fixtures", "packages")


class PackageBaseMock(PackageBase):
    def _get_directory_node(self):
        pass

    def _load_from_directory_node(self, directory_node):
        pass


class TestPackageBase(unittest.TestCase):
    def setUp(self):
        self.p1 = PackageBaseMock("http://urlA", "projA", "pkgA")

    def test_str(self):
        self.assertEqual(str(self.p1), "projA/pkgA")

    def test_repr(self):
        self.assertTrue(repr(self.p1).endswith("(projA/pkgA)"))

    def test_eq(self):
        # the same
        p2 = PackageBaseMock(self.p1.apiurl, self.p1.project, self.p1.name)
        self.assertEqual(self.p1, p2)

        # package name differs
        p2 = PackageBaseMock(self.p1.apiurl, self.p1.project, "pkgB")
        self.assertNotEqual(self.p1, p2)

        # project name differs
        p2 = PackageBaseMock(self.p1.apiurl, "projB", self.p1.name)
        self.assertNotEqual(self.p1, p2)

        # baseurl differs
        p2 = PackageBaseMock("http://urlB", self.p1.project, self.p1.name)
        self.assertNotEqual(self.p1, p2)

    def test_lt(self):
        # the same
        p2 = PackageBaseMock(self.p1.apiurl, self.p1.project, self.p1.name)
        self.assertFalse(self.p1 < p2)

        # package name differs
        p2 = PackageBaseMock(self.p1.apiurl, self.p1.project, "pkgB")
        self.assertTrue(self.p1 < p2)

        # project name differs
        p2 = PackageBaseMock(self.p1.apiurl, "projB", self.p1.name)
        self.assertTrue(self.p1 < p2)

        # baseurl differs
        p2 = PackageBaseMock("http://urlB", self.p1.project, self.p1.name)
        self.assertTrue(self.p1 < p2)

    def test_hash(self):
        p2 = PackageBaseMock(self.p1.apiurl, self.p1.project, self.p1.name)
        self.assertEqual(hash(self.p1), hash(p2))

        packages = set()
        packages.add(self.p1)
        # the second instance appears to be there because it has the same hash
        # it is ok, because we consider such packages equal
        self.assertIn(p2, packages)


class TestLocalPackage(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def test_load(self):
        path = os.path.join(self.tmpdir, "osctest", "openSUSE:Tools", "osc")
        p = LocalPackage(path)

        self.assertEqual(p.name, "osc")
        self.assertEqual(p.project, "openSUSE:Tools")
        self.assertEqual(p.apiurl, "http://localhost")
        self.assertEqual(p.rev, "373")
        self.assertEqual(p.vrev, "339")
        self.assertEqual(p.srcmd5, "30ccce6c3a1a4322e79c2935a52af18b")

        self.assertEqual(p.linkinfo.project, "openSUSE:Factory")
        self.assertEqual(p.linkinfo.package, "osc")
        self.assertEqual(p.linkinfo.srcmd5, "1ccbcd1b0b531a37ad75b34b5a1e2e3e")
        self.assertEqual(p.linkinfo.baserev, "2c3ae65909d69e0f63113ccfe0e5f3f8")
        self.assertEqual(p.linkinfo.xsrcmd5, "6a31b956f9431b0644ad6cf8e845c4e5")
        self.assertEqual(p.linkinfo.lsrcmd5, "30ccce6c3a1a4322e79c2935a52af18b")

        self.assertEqual(len(p.files), 3)

        f = p.files[0]
        self.assertEqual(f.name, "osc-0.182.0.tar.gz")
        self.assertEqual(f.md5, "87f040c76f3da86fd7218c972b9df1dc")
        self.assertEqual(f.size, 381596)
        self.assertEqual(f.mtime, 1662638726)


class TestApiPackage(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    @GET("http://localhost/source/openSUSE:Tools/osc", file="osctest/openSUSE:Tools/osc/.osc/_files")
    def test_load(self):
        p = ApiPackage("http://localhost", "openSUSE:Tools", "osc")

        self.assertEqual(p.name, "osc")
        self.assertEqual(p.project, "openSUSE:Tools")
        self.assertEqual(p.apiurl, "http://localhost")
        self.assertEqual(p.rev, "373")
        self.assertEqual(p.vrev, "339")
        self.assertEqual(p.srcmd5, "30ccce6c3a1a4322e79c2935a52af18b")

        self.assertEqual(p.linkinfo.project, "openSUSE:Factory")
        self.assertEqual(p.linkinfo.package, "osc")
        self.assertEqual(p.linkinfo.srcmd5, "1ccbcd1b0b531a37ad75b34b5a1e2e3e")
        self.assertEqual(p.linkinfo.baserev, "2c3ae65909d69e0f63113ccfe0e5f3f8")
        self.assertEqual(p.linkinfo.xsrcmd5, "6a31b956f9431b0644ad6cf8e845c4e5")
        self.assertEqual(p.linkinfo.lsrcmd5, "30ccce6c3a1a4322e79c2935a52af18b")

        self.assertEqual(len(p.files), 3)

        f = p.files[0]
        self.assertEqual(f.name, "osc-0.182.0.tar.gz")
        self.assertEqual(f.md5, "87f040c76f3da86fd7218c972b9df1dc")
        self.assertEqual(f.size, 381596)
        self.assertEqual(f.mtime, 1662638726)


if __name__ == "__main__":
    unittest.main()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import

from .enums import ReleaseTriggers


class RepositoryReleasetarget(XmlModel):
    XML_TAG = "releasetarget"

    project: str = Field(
        xml_attribute=True,
    )

    repository: str = Field(
        xml_attribute=True,
    )

    trigger: Optional[ReleaseTriggers] = Field(
        xml_attribute=True,
    )



from . import package as osc_package


def forward_request(apiurl, request, interactive=True):
    """
    Forward the specified `request` to the projects the packages were branched from.
    """
    from .. import core as osc_core

    for action in request.get_actions("submit"):
        package = osc_package.ApiPackage(apiurl, action.tgt_project, action.tgt_package)

        if not package.linkinfo:
            # not a linked/branched package, can't forward to parent
            continue

        project = package.linkinfo.project
        package = package.linkinfo.package

        if interactive:
            reply = input(f"\nForward request to {project}/{package}? ([y]/n) ")
            if reply.lower() not in ("y", ""):
                continue

        msg = f"Forwarded request #{request.reqid} from {request.creator}\n\n{request.description}"
        new_request_id = osc_core.create_submit_request(
            apiurl,
            action.tgt_project,
            action.tgt_package,
            project,
            package,
            msg,
        )
        msg = f"Forwarded request #{request.reqid} from {request.creator} to {project}/{package}: #{new_request_id}"
        print(msg)



from typing import Optional

from .connection import Connection
from .connection import GiteaHTTPResponse
from .exceptions import BranchDoesNotExist
from .exceptions import BranchExists
from .exceptions import GiteaException


class Branch:
    @classmethod
    def get(
        cls,
        conn: Connection,
        owner: str,
        repo: str,
        branch: str,
    ) -> GiteaHTTPResponse:
        """
        Retrieve details about a repository branch.

        :param conn: Gitea ``Connection`` instance.
        :param owner: Owner of the repo.
        :param repo: Name of the repo.
        :param branch: Name of the branch.
        """
        url = conn.makeurl("repos", owner, repo, "branches", branch)
        try:
            return conn.request("GET", url)
        except GiteaException as e:
            if e.status == 404:
                raise BranchDoesNotExist(e.response, owner, repo, branch) from None
            raise

    @classmethod
    def list(
        cls,
        conn: Connection,
        owner: str,
        repo: str,
    ) -> GiteaHTTPResponse:
        """
        Retrieve details about all repository branches.

        :param conn: Gitea ``Connection`` instance.
        :param owner: Owner of the repo.
        :param repo: Name of the repo.
        """
        url = conn.makeurl("repos", owner, repo, "branches")
        # XXX: returns 'null' when there are no branches; an empty list would be a better API
        return conn.request("GET", url)

    @classmethod
    def create(
        cls,
        conn: Connection,
        owner: str,
        repo: str,
        *,
        old_ref_name: Optional[str] = None,
        new_branch_name: str,
        exist_ok: bool = False,
    ) -> GiteaHTTPResponse:
        """
        Create a new branch in a repository.

        :param conn: Gitea ``Connection`` instance.
        :param owner: Owner of the repo.
        :param repo: Name of the repo.
        :param old_ref_name: Name of the old branch/tag/commit to create from.
        :param new_branch_name: Name of the branch to create.
        :param exist_ok: A ``BranchExists`` exception is raised when the target exists. Set to ``True`` to avoid throwing the exception.
        """
        json_data = {
            "new_branch_name": new_branch_name,
            "old_ref_name": old_ref_name,
        }
        url = conn.makeurl("repos", owner, repo, "branches")
        try:
            return conn.request("POST", url, json_data=json_data)
        except GiteaException as e:
            if e.status == 409:
                if exist_ok:
                    return cls.get(conn, owner, repo, new_branch_name)
                raise BranchExists(e.response, owner, repo, new_branch_name) from None
            raise



import unittest

import osc.conf
from osc.build import check_trusted_projects
from osc.oscerr import UserAbort


class TestTrustedProjects(unittest.TestCase):
    def setUp(self):
        osc.conf.config = osc.conf.Options()

    def test_name(self):
        apiurl = "https://example.com"
        osc.conf.config["apiurl"] = apiurl
        osc.conf.config.setdefault("api_host_options", {}).setdefault(apiurl, {}).setdefault("trusted_prj", None)

        osc.conf.config["api_host_options"][apiurl]["trusted_prj"] = []
        self.assertRaises(UserAbort, check_trusted_projects, apiurl, ["foo"], interactive=False)

        osc.conf.config["api_host_options"][apiurl]["trusted_prj"] = ["qwerty", "foo", "asdfg"]
        check_trusted_projects(apiurl, ["foo"], interactive=False)

    def test_glob(self):
        apiurl = "https://example.com"
        osc.conf.config["apiurl"] = apiurl
        osc.conf.config.setdefault("api_host_options", {}).setdefault(apiurl, {}).setdefault("trusted_prj", None)

        osc.conf.config["api_host_options"][apiurl]["trusted_prj"] = ["f*"]
        check_trusted_projects(apiurl, ["foo"], interactive=False)


if __name__ == "__main__":
    unittest.main()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import

from .enums import LocalRole


class GroupRole(XmlModel):
    XML_TAG = "group"

    groupid: str = Field(
        xml_attribute=True,
    )

    role: LocalRole = Field(
        xml_attribute=True,
    )



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RequestActionTarget(XmlModel):
    XML_TAG = "target"

    project: str = Field(
        xml_attribute=True,
    )

    package: Optional[str] = Field(
        xml_attribute=True,
    )

    releaseproject: Optional[str] = Field(
        xml_attribute=True,
    )

    repository: Optional[str] = Field(
        xml_attribute=True,
    )



import sys
import textwrap
from typing import Dict
from typing import Optional

from .. import oscerr
from .tty import colorize


def get_user_input(question: str, answers: Dict[str, str], default_answer: Optional[str] = None) -> str:
    """
    Ask user a question and wait for reply.

    :param question: The question. The text gets automatically dedented and stripped.
    :param answers: A dictionary with answers. Keys are the expected replies and values are their descriptions.
    :param default_answer: The default answer. Must be ``None`` or match an ``answers`` entry.
    """

    if default_answer and default_answer not in answers:
        raise ValueError(f"Default answer doesn't match any answer: {default_answer}")

    question = textwrap.dedent(question)
    question = question.strip()

    prompt = []
    for key, value in answers.items():
        value = f"{colorize(key, 'bold')}){value}"
        prompt.append(value)

    prompt_str = " / ".join(prompt)
    if default_answer:
        prompt_str += f" (default={colorize(default_answer, 'bold')})"
    prompt_str += ": "

    print(question, file=sys.stderr)

    while True:
        try:
            reply = input(prompt_str)
        except EOFError:
            # interpret ctrl-d as user abort
            raise oscerr.UserAbort()  # pylint: disable=raise-missing-from

        if reply in answers:
            return reply
        if reply.strip() in answers:
            return reply.strip()
        if not reply.strip():
            return default_answer

        print(f"Invalid reply: {colorize(reply, 'bold,red')}", file=sys.stderr)



class Linkinfo:
    """linkinfo metadata (which is part of the xml representing a directory)
    """

    def __init__(self):
        """creates an empty linkinfo instance"""
        self.project = None
        self.package = None
        self.xsrcmd5 = None
        self.lsrcmd5 = None
        self.srcmd5 = None
        self.error = None
        self.rev = None
        self.baserev = None

    def read(self, linkinfo_node):
        """read in the linkinfo metadata from the ``<linkinfo>`` element passed as
        elementtree node.
        If the passed element is ``None``, the method does nothing.
        """
        if linkinfo_node is None:
            return
        self.project = linkinfo_node.get('project')
        self.package = linkinfo_node.get('package')
        self.xsrcmd5 = linkinfo_node.get('xsrcmd5')
        self.lsrcmd5 = linkinfo_node.get('lsrcmd5')
        self.srcmd5 = linkinfo_node.get('srcmd5')
        self.error = linkinfo_node.get('error')
        self.rev = linkinfo_node.get('rev')
        self.baserev = linkinfo_node.get('baserev')

    def islink(self):
        """:return: ``True`` if the linkinfo is not empty, otherwise ``False``"""
        if self.xsrcmd5 or self.lsrcmd5 or self.error is not None:
            return True
        return False

    def isexpanded(self):
        """:return: ``True`` if the package is an expanded link"""
        if self.lsrcmd5 and not self.xsrcmd5:
            return True
        return False

    def haserror(self):
        """:return: ``True`` if the link is in error state (could not be applied)"""
        if self.error:
            return True
        return False

    def __str__(self):
        """return an informatory string representation"""
        if self.islink() and not self.isexpanded():
            return 'project %s, package %s, xsrcmd5 %s, rev %s' \
                % (self.project, self.package, self.xsrcmd5, self.rev)
        elif self.islink() and self.isexpanded():
            if self.haserror():
                return 'broken link to project %s, package %s, srcmd5 %s, lsrcmd5 %s: %s' \
                    % (self.project, self.package, self.srcmd5, self.lsrcmd5, self.error)
            else:
                return 'expanded link to project %s, package %s, srcmd5 %s, lsrcmd5 %s' \
                    % (self.project, self.package, self.srcmd5, self.lsrcmd5)
        else:
            return 'None'



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class BlockModes(str, Enum):
    ALL = "all"
    LOCAL = "local"
    NEVER = "never"


class BoolString(str, Enum):
    TRUE = "true"
    FALSE = "false"


class BuildArch(str, Enum):
    NOARCH = "noarch"
    AARCH64 = "aarch64"
    AARCH64_ILP32 = "aarch64_ilp32"
    ARMV4L = "armv4l"
    ARMV5L = "armv5l"
    ARMV6L = "armv6l"
    ARMV7L = "armv7l"
    ARMV5EL = "armv5el"
    ARMV6EL = "armv6el"
    ARMV7EL = "armv7el"
    ARMV7HL = "armv7hl"
    ARMV8EL = "armv8el"
    HPPA = "hppa"
    M68K = "m68k"
    I386 = "i386"
    I486 = "i486"
    I586 = "i586"
    I686 = "i686"
    ATHLON = "athlon"
    IA64 = "ia64"
    K1OM = "k1om"
    LOONGARCH64 = "loongarch64"
    MIPS = "mips"
    MIPSEL = "mipsel"
    MIPS32 = "mips32"
    MIPS64 = "mips64"
    MIPS64EL = "mips64el"
    PPC = "ppc"
    PPC64 = "ppc64"
    PPC64P7 = "ppc64p7"
    PPC64LE = "ppc64le"
    RISCV64 = "riscv64"
    S390 = "s390"
    S390X = "s390x"
    SH4 = "sh4"
    SPARC = "sparc"
    SPARC64 = "sparc64"
    SPARC64V = "sparc64v"
    SPARCV8 = "sparcv8"
    SPARCV9 = "sparcv9"
    SPARCV9V = "sparcv9v"
    X86_64 = "x86_64"
    LOCAL = "local"


class LinkedbuildModes(str, Enum):
    OFF = "off"
    LOCALDEP = "localdep"
    ALLDIRECT = "alldirect"
    ALL = "all"


class LocalRole(str, Enum):
    MAINTAINER = "maintainer"
    BUGOWNER = "bugowner"
    REVIEWER = "reviewer"
    DOWNLOADER = "downloader"
    READER = "reader"


class ObsRatings(str, Enum):
    LOW = "low"
    MODERATE = "moderate"
    IMPORTANT = "important"
    CRITICAL = "critical"


class RebuildModes(str, Enum):
    TRANSITIVE = "transitive"
    DIRECT = "direct"
    LOCAL = "local"


class ReleaseTriggers(str, Enum):
    MANUAL = "manual"
    MAINTENANCE = "maintenance"
    OBSGENDIFF = "obsgendiff"


class RequestStates(str, Enum):
    REVIEW = "review"
    NEW = "new"
    ACCEPTED = "accepted"
    DECLINED = "declined"
    REVOKED = "revoked"
    SUPERSEDED = "superseded"
    DELETED = "deleted"



import sys

import osc.commandline_git


def b(value: str):
    if value is not None:
        return value.lower() in ["1", "yes", "true", "on"]
    return None


class PullRequestSetCommand(osc.commandline_git.GitObsCommand):
    """
    Change a pull request
    """

    name = "set"
    parent = "PullRequestCommand"

    def init_arguments(self):
        self.add_argument_owner_repo_pull(nargs="+")
        self.add_argument(
            "--title",
        )
        self.add_argument(
            "--description",
        )
        self.add_argument(
            "--allow-maintainer-edit",
            action=osc.commandline_git.BooleanAction,
            help="Users with write access to the base branch can also push to the pull request's head branch",
        )

    def run(self, args):
        from osc import gitea_api
        from osc.core import highlight_diff
        from osc.output import tty

        self.print_gitea_settings()
        print(args)

        num_entries = 0
        failed_entries = []
        for owner, repo, pull in args.owner_repo_pull:
            try:
                pr = gitea_api.PullRequest.set(
                    self.gitea_conn,
                    owner,
                    repo,
                    int(pull),
                    title=args.title,
                    description=args.description,
                    allow_maintainer_edit=args.allow_maintainer_edit,
                ).json()
                num_entries += 1
            except gitea_api.GiteaException as e:
                if e.status == 404:
                    failed_entries.append(f"{owner}/{repo}#{pull}")
                    continue
                raise

            print(gitea_api.PullRequest.to_human_readable_string(pr))
            print()

        print(f"Total modified entries: {num_entries}", file=sys.stderr)
        if failed_entries:
            print(
                f"{tty.colorize('ERROR', 'red,bold')}: Couldn't change the following pull requests: {', '.join(failed_entries)}",
                file=sys.stderr,
            )
            sys.exit(1)



import os
import tempfile
import unittest

import osc.conf
import osc.grabber as osc_grabber


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "conf_fixtures")


class TestMirrorGroup(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='osc_test')
        oscrc = os.path.join(self._get_fixtures_dir(), "oscrc")
        osc.conf.get_config(override_conffile=oscrc, override_no_keyring=True)

    def tearDown(self):
        try:
            shutil.rmtree(self.tmpdir)
        except:
            pass

    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def test_invalid_scheme(self):
        gr = osc_grabber.OscFileGrabber()
        mg = osc_grabber.OscMirrorGroup(gr, ["container://example.com"])
        mg.urlgrab(None, os.path.join(self.tmpdir, "file"))


if __name__ == "__main__":
    unittest.main()



import textwrap

from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .keyinfo_pubkey import KeyinfoPubkey
from .keyinfo_sslcert import KeyinfoSslcert


class Keyinfo(XmlModel):
    XML_TAG = "keyinfo"

    project: Optional[str] = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            The name of the project.
            """
        ),
    )

    pubkey_list: Optional[List[KeyinfoPubkey]] = Field(
        xml_name="pubkey",
    )

    sslcert_list: Optional[List[KeyinfoSslcert]] = Field(
        xml_name="sslcert",
    )

    @classmethod
    def from_api(cls, apiurl: str, project: str) -> "Keyinfo":
        url_path = ["source", project, "_keyinfo"]
        url_query = {}
        response = cls.xml_request("GET", apiurl, url_path, url_query)
        return cls.from_file(response, apiurl=apiurl)


    @classmethod
    def get_pubkey_deprecated(cls, apiurl: str, project: str, *, traverse: bool = True) -> Optional[Tuple[str, str]]:
        """
        Old API for retrieving pubkey of the given ``project``. Use ``Keyinfo.from_api()`` instead if possible.

        :param traverse: If set to ``True`` and the key is not found, traverse project hierarchy for the first available key.
        :return: (project, pubkey) or None
        """
        from urllib.error import HTTPError
        from ..connection import http_request
        from ..core import makeurl
        from ..output import print_msg

        while True:
            url_path = ["source", project, "_pubkey"]
            url_query = {}
            url = makeurl(apiurl, url_path, url_query)
            try:
                response = http_request("GET", url)
                pubkey = response.read().decode("utf-8")
                return project, pubkey
            except HTTPError as e:
                if e.code != 404:
                    raise

                if not traverse:
                    return None

                parts = project.rsplit(":", 1)
                if parts[0] != project:
                    print_msg(f"No pubkey found in project '{project}'. Trying the parent project '{parts[0]}'...", print_to="debug")
                    project = parts[0]
                    continue

                # we're at the top level, no key found
                return None

    @classmethod
    def get_sslcert_deprecated(cls, apiurl: str, project: str, *, traverse: bool = True) -> Optional[Tuple[str, str]]:
        """
        Old API for retrieving sslcert of the given ``project``. Use ``Keyinfo.from_api()`` instead if possible.

        :param traverse: If set to ``True`` and the cert is not found, traverse project hierarchy for the first available cert.
        :return: (project, sslcert) or None
        """
        from urllib.error import HTTPError
        from ..connection import http_request
        from ..core import makeurl
        from ..output import print_msg

        while True:
            url_path = ["source", project, "_project", "_sslcert"]
            url_query = {
                "meta": 1,
            }
            url = makeurl(apiurl, url_path, url_query)
            try:
                response = http_request("GET", url)
                sslcert = response.read().decode("utf-8")
                return project, sslcert
            except HTTPError as e:
                if e.code != 404:
                    raise

                if not traverse:
                    return None

                parts = project.rsplit(":", 1)
                if parts[0] != project:
                    print_msg(f"No sslcert found in project '{project}'. Trying the parent project '{parts[0]}'...", print_to="debug")
                    project = parts[0]
                    continue

                # we're at the top level, no cert found
                return None
